"""

Tools for visualizing annotations on a Document's pages.

@kylel

"""

from typing import Dict, List, Optional, Union

import layoutparser.elements as lpe
import layoutparser.visualization as lpv

from papermage.magelib import Entity, Image


def plot_entities_on_page(
    page_image: Image,
    entities: List[Entity],
    box_width: Optional[Union[List[int], int]] = None,
    box_alpha: Optional[Union[List[float], float]] = None,
    box_color: Optional[Union[List[str], str]] = None,
    color_map: Optional[Dict] = None,
    show_element_id: bool = False,
    show_element_type: bool = False,
    id_font_size: Optional[int] = None,
    id_font_path: Optional[str] = None,
    id_text_color: Optional[str] = None,
    id_text_background_color: Optional[str] = None,
    id_text_background_alpha: Optional[float] = 1,
) -> Image:
    page_w, page_h = page_image.pilimage.size
    lpe_boxes = []
    for entity in entities:
        for box in entity.boxes:
            lpe_box = lpe.Rectangle(
                x_1=box.l * page_w, y_1=box.t * page_h, x_2=(box.l + box.w) * page_w, y_2=(box.t + box.h) * page_h
            )
            lpe_boxes.append(lpe_box)
    viz = lpv.draw_box(
        canvas=page_image.pilimage,
        layout=lpe_boxes,
        box_width=box_width,
        box_alpha=box_alpha,
        box_color=box_color,
        color_map=color_map,
        show_element_id=show_element_id,
        show_element_type=show_element_type,
        id_font_size=id_font_size,
        id_font_path=id_font_path,
        id_text_color=id_text_color,
        id_text_background_color=id_text_background_color,
        id_text_background_alpha=id_text_background_alpha,
    )
    # return viz as new image
    annotated_image = Image()
    annotated_image.pilimage = viz
    return annotated_image



import json
from typing import List

from decontext.data_types import (
    EvidenceParagraph,
    PaperContext,
    PaperSnippet,
    QuestionAnswerEvidence,
)
from decontext.step.qa import TemplateRetrievalQAStep

from papermage.magelib import Annotation, Document, Entity
from papermage.predictors.base_predictor import BasePredictor


class APISpanQAPredictor(BasePredictor):
    REQUIRED_BACKENDS = ["transformers", "torch", "decontext"]

    def __init__(self, context_unit_name: str = "paragraph", span_name: str = "user_selected_span"):
        self.context_unit_name = context_unit_name
        self.span_name = span_name

        self.retrieval_qa_step = TemplateRetrievalQAStep()

    @property
    def REQUIRED_DOCUMENT_FIELDS(self) -> List[str]:
        return [self.context_unit_name, self.span_name]

    def preprocess(self, doc: Document) -> PaperSnippet:
        paper_context = PaperContext.parse_raw(
            json.dumps(
                {
                    "title": getattr(doc, "title") if "title" in doc.fields else "",
                    "abstract": getattr(doc, "abstract") if "abstract" in doc.fields else "",
                    "full_text": [
                        {
                            "section_name": "",
                            "paragraphs": [entity.text for entity in getattr(doc, self.context_unit_name)],
                        }
                    ],
                }
            )
        )

        context = getattr(getattr(doc, self.span_name)[0], self.context_unit_name)[0]
        # Get the index of the context with the span
        if context.id is not None:
            context_index = context.id
        else:
            context_index = [i for i, c in enumerate(doc.get_entity(self.context_unit_name)) if c == context][0]

        paper_snippet = PaperSnippet(
            snippet=getattr(doc, self.span_name)[0].text,
            context=paper_context,
            qae=[],
            paragraph_with_snippet=EvidenceParagraph(
                section="",
                paragraph=context.text,
                index=context_index,
            ),
        )

        paper_snippet.add_question(getattr(doc, self.span_name)[0].metadata["question"])

        return paper_snippet

    def _predict(self, doc: Document) -> List[Annotation]:
        # preprocess to format this for decontext
        paper_snippet = self.preprocess(doc)

        # run the decontext qa step
        self.retrieval_qa_step.run(paper_snippet)

        # postprocess to format back into papermage doc format
        annotations: List[Annotation] = []

        # add the question and answer to a copy of the user selected span
        user_selected_span = getattr(doc, self.span_name)
        new_user_selected_span = Entity.from_json(user_selected_span[0].to_json())
        new_user_selected_span.metadata["type"] = "answer"
        new_user_selected_span.metadata["question"] = paper_snippet.qae[0].question
        new_user_selected_span.metadata["answer"] = paper_snippet.qae[0].answer
        annotations.append(new_user_selected_span)
        
        # add the context with span
        context_with_span = getattr(getattr(doc, self.span_name)[0], self.context_unit_name)[0]
        # context_with_span = doc.get_entity(self.context_unit_name)[paper_snippet.paragraph_with_snippet.index]
        new_context_with_span = Entity.from_json(context_with_span.to_json())
        new_context_with_span.metadata["type"] = "context_with_span"
        annotations.append(new_context_with_span)

        # add the evidence entities
        for evidence in paper_snippet.qae[0].evidence:
            entity = doc.get_entity(self.context_unit_name)[evidence.index]
            new_entity = Entity.from_json(entity.to_json())
            new_entity.metadata["type"] = "evidence"
            annotations.append(new_entity)

        return annotations



"""

@benjaminn, @kylel

"""

import json
import os
import pathlib
import re
import unittest

from papermage.magelib import Document, Entity, Image, Span
from papermage.rasterizers import PDF2ImageRasterizer


class TestPDF2ImageRasterizer(unittest.TestCase):
    def setUp(cls) -> None:
        cls.fixture_path = pathlib.Path(__file__).parent.parent / "fixtures"

    def test_raseterize(self):
        rasterizer = PDF2ImageRasterizer()
        images = rasterizer.rasterize(input_pdf_path=self.fixture_path / "2304.02623v1.pdf", dpi=72)
        assert len(images) == 4
        assert images[0].pilimage.size == (612, 792)
        assert isinstance(images[0], Image)

    def test_attach(self):
        rasterizer = PDF2ImageRasterizer()
        image = Image.create_rgb_random()
        doc = Document(symbols="This is a doc.")
        page = Entity(spans=[Span(start=0, end=14)])
        doc.annotate_entity(field_name="pages", entities=[page])
        rasterizer.attach_images(images=[image], doc=doc)
        assert doc.pages[0].images[0] == image



"""

Test sentence predictor

@kylel

"""

import json
import pathlib
import unittest

from papermage.magelib import Document, Entity, Span
from papermage.predictors import PysbdSentencePredictor, SVMWordPredictor


class TestPysbdSentencePredictor(unittest.TestCase):
    def setUp(self):
        self.fixture_path = pathlib.Path(__file__).parent.parent / "fixtures"
        with open(self.fixture_path / "2304.02623v1.json") as f_in:
            self.doc = Document.from_json(json.load(f_in))
        # TODO: seems like Words causes SentencePredictor to behave better
        #       may be worth testing.
        # word_predictor = SVMWordPredictor.from_path(
        #     tar_path=str(self.fixture_path / "svm_word_predictor/svm_word_predictor.tar.gz")
        # )
        # words = word_predictor.predict(self.doc)
        # self.doc.annotate_entity(field_name="words", entities=words)

    def test_predict(self):
        predictor = PysbdSentencePredictor()
        doc = Document(symbols="This is a test. This is another test.")
        tokens = [
            Entity(spans=[Span(0, 4)]),
            Entity(spans=[Span(5, 7)]),
            Entity(spans=[Span(8, 9)]),
            Entity(spans=[Span(10, 14)]),
            Entity(spans=[Span(14, 15)]),
            Entity(spans=[Span(16, 20)]),
            Entity(spans=[Span(21, 23)]),
            Entity(spans=[Span(24, 31)]),
            Entity(spans=[Span(32, 36)]),
            Entity(spans=[Span(36, 37)]),
        ]
        doc.annotate_entity(field_name="tokens", entities=tokens)
        sents = predictor.predict(doc)
        doc.annotate_entity(field_name="sentences", entities=sents)
        self.assertEqual(len(doc.sentences), 2)
        self.assertEqual(doc.sentences[0].text, "This is a test.")
        self.assertEqual(doc.sentences[1].text, "This is another test.")

    def test_predict_paper(self):
        predictor = PysbdSentencePredictor()
        sents = predictor.predict(self.doc)
        self.doc.annotate_entity(field_name="sentences", entities=sents)
        self.assertEqual(len(self.doc.sentences), 310)

        valid_sent_1 = """Large language models have introduced exciting new opportunities and challenges in designing and developing new AI-assisted writing support tools."""
        self.assertTrue(any([valid_sent_1 in s.text.replace("\n", " ") for s in self.doc.sentences[:10]]))

        valid_sent_2 = """Recent work has shown that leveraging this new tech- nology can transform writing in many scenarios such as ideation during creative writing, editing support, and summarization."""
        self.assertTrue(any([valid_sent_2 in s.text.replace("\n", " ") for s in self.doc.sentences[:10]]))

        valid_sent_3 = """How- ever, AI-supported expository writing —including real-world tasks like scholars writing literature reviews or doctors writing progress notes—is relatively understudied."""
        self.assertTrue(any([valid_sent_3 in s.text.replace("\n", " ") for s in self.doc.sentences[:10]]))

        valid_sent_4 = """In this position paper, we argue that developing AI supports for expository writing has unique and exciting research challenges and can lead to high real-world impacts."""
        self.assertTrue(any([valid_sent_4 in s.text.replace("\n", " ") for s in self.doc.sentences[:10]]))



"""

@kylel, benjaminn
"""

import json
import os
import pathlib
import re
import unittest

import numpy as np

from papermage.magelib import Box, Document, Entity, Span
from papermage.parsers import PDFPlumberParser


class TestPDFPlumberParser(unittest.TestCase):
    def setUp(self) -> None:
        self.fixture_path = pathlib.Path(__file__).parent.parent / "fixtures"
        self.parser = PDFPlumberParser()
        self.doc = self.parser.parse(input_pdf_path=str(self.fixture_path / "2304.02623v1.pdf"))

    def test_parse(self):
        # right output type
        assert isinstance(self.doc, Document)
        # the right fields
        assert self.doc.symbols
        assert self.doc.pages
        assert self.doc.tokens
        assert self.doc.rows
        # roughly the right content
        for keyword in [
            "Beyond",
            "Summarization",
            "Designing",
            "AI",
            "Support",
            "Real",
            "World",
            "Expository",
            "Writing",
            "Tasks",
        ]:
            assert keyword in self.doc.symbols[:100]

    def test_parse_page_dims(self):
        for page in self.doc.pages:
            self.assertEqual(612.0, page.metadata.width)
            self.assertEqual(792.0, page.metadata.height)
            self.assertEqual(1.0, page.metadata.user_unit)

    def test_non_default_user_unit(self):
        doc = self.parser.parse(input_pdf_path=self.fixture_path / "test-uu.pdf")
        for page in doc.pages:
            self.assertEqual(595.0, page.metadata.width)
            self.assertEqual(842.0, page.metadata.height)
            self.assertEqual(2.0, page.metadata.user_unit)

    def test_parse_fontinfo(self):
        metadata = self.doc.tokens[0].metadata  # pylint: disable=no-member
        self.assertEqual("UCCPKS+LinBiolinumTB", metadata["fontname"])
        self.assertAlmostEqual(17.2154, metadata["size"])

    def test_split_punctuation(self):
        no_split_parser = PDFPlumberParser(split_at_punctuation=False)
        no_split_doc = no_split_parser.parse(input_pdf_path=self.fixture_path / "2304.02623v1.pdf")
        no_split_tokens_with_numbers = [
            token.text for token in no_split_doc.tokens if re.search(r"[0-9]", token.text)
        ]
        assert "[24]." in no_split_tokens_with_numbers
        assert "31]," in no_split_tokens_with_numbers

        custom_split_parser = PDFPlumberParser(split_at_punctuation=",.[]:")
        custom_split_doc = custom_split_parser.parse(input_pdf_path=self.fixture_path / "2304.02623v1.pdf")
        custom_split_tokens_with_numbers = [
            token.text for token in custom_split_doc.tokens if re.search(r"[0-9]", token.text)
        ]
        assert "[24]." not in custom_split_tokens_with_numbers
        assert "31]," not in custom_split_tokens_with_numbers
        assert "(2019)" in custom_split_tokens_with_numbers

        default_split_parser = PDFPlumberParser(split_at_punctuation=True)
        default_split_doc = default_split_parser.parse(input_pdf_path=self.fixture_path / "2304.02623v1.pdf")
        default_split_tokens_with_numbers = [
            token.text for token in default_split_doc.tokens if re.search(r"[0-9]", token.text)
        ]
        assert "(2019)" not in default_split_tokens_with_numbers

        assert (
            len(no_split_tokens_with_numbers)
            < len(custom_split_tokens_with_numbers)
            < len(default_split_tokens_with_numbers)
        )

    def test_align_coarse_and_fine_tokens(self):
        # example
        coarse_tokens = ["abc", "def"]
        fine_tokens = ["ab", "c", "d", "ef"]
        out = self.parser._align_coarse_and_fine_tokens(coarse_tokens=coarse_tokens, fine_tokens=fine_tokens)
        assert out == [0, 0, 1, 1]

        # minimal case
        coarse_tokens = []
        fine_tokens = []
        out = self.parser._align_coarse_and_fine_tokens(coarse_tokens=coarse_tokens, fine_tokens=fine_tokens)
        assert out == []

        # identical case
        coarse_tokens = ["a", "b", "c"]
        fine_tokens = ["a", "b", "c"]
        out = self.parser._align_coarse_and_fine_tokens(coarse_tokens=coarse_tokens, fine_tokens=fine_tokens)
        assert out == [0, 1, 2]

        # misaligned case
        with self.assertRaises(AssertionError):
            coarse_tokens = ["a", "b"]
            fine_tokens = ["ab"]
            self.parser._align_coarse_and_fine_tokens(coarse_tokens=coarse_tokens, fine_tokens=fine_tokens)

        # same num of chars, but chars mismatch case
        with self.assertRaises(AssertionError):
            coarse_tokens = ["ab"]
            fine_tokens = ["a", "c"]
            self.parser._align_coarse_and_fine_tokens(coarse_tokens=coarse_tokens, fine_tokens=fine_tokens)

    def test_convert_nested_text_to_doc_json(self):
        # example
        token_dicts = [
            {"text": text, "bbox": Box(l=0.0, t=0.1, w=0.2, h=0.3, page=4)}
            for text in ["ab", "c", "d", "ef", "gh", "i", "j", "kl"]
        ]
        word_ids = [0, 0, 1, 2, 3, 4, 5, 5]
        row_ids = [0, 0, 1, 1, 2, 2, 3, 3]
        page_ids = [0, 0, 0, 0, 1, 1, 1, 1]
        page_dims = [(100, 200, 1.0), (400, 800, 1.0)]
        out = self.parser._convert_nested_text_to_doc_json(
            token_dicts=token_dicts,
            word_ids=word_ids,
            row_ids=row_ids,
            page_ids=page_ids,
            dims=page_dims,
        )
        assert out["symbols"] == "abc\nd ef\ngh i\njkl"
        tokens = [Entity.from_json(entity_json=t_dict) for t_dict in out["entities"]["tokens"]]
        assert [(t.start, t.end) for t in tokens] == [
            (0, 2),
            (2, 3),
            (4, 5),
            (6, 8),
            (9, 11),
            (12, 13),
            (14, 15),
            (15, 17),
        ]
        assert [out["symbols"][t.start : t.end] for t in tokens] == [
            "ab",
            "c",
            "d",
            "ef",
            "gh",
            "i",
            "j",
            "kl",
        ]
        rows = [Entity.from_json(entity_json=r_dict) for r_dict in out["entities"]["rows"]]
        assert [(r.start, r.end) for r in rows] == [(0, 3), (4, 8), (9, 13), (14, 17)]
        assert [out["symbols"][r.start : r.end] for r in rows] == [
            "abc",
            "d ef",
            "gh i",
            "jkl",
        ]
        pages = [Entity.from_json(entity_json=p_dict) for p_dict in out["entities"]["pages"]]
        assert [(p.start, p.end) for p in pages] == [(0, 8), (9, 17)]
        assert [out["symbols"][p.start : p.end] for p in pages] == [
            "abc\nd ef",
            "gh i\njkl",
        ]

    def test_parser_stability(self):
        """
        We need output to be stable from release to release. Failure of this test is caused
        by changes to core output: document text, tokenization, and bbox localization.
        It deliberately excludes `metadata` from consideration as we are expanding
        its scope of coverage, but that should probably be locked down too the moment
        we depend on particular fields.

        Updates that break this test should be considered potentially breaking to downstream
        models and require re-evaluation and possibly retraining of all components in the DAG.
        """
        with open(self.fixture_path / "2304.02623v1.json", "r") as f:
            raw_json = f.read()
            fixture_doc_json = json.loads(raw_json)
            fixture_doc = Document.from_json(fixture_doc_json)

        self.assertEqual(
            self.doc.symbols, fixture_doc.symbols, msg="Current parse has extracted different text from pdf."
        )

        def compare_entities(current_doc_ents, fixture_doc_ents, annotation_name):
            current_doc_ents_simplified = [[(s.start, s.end) for s in ent.spans] for ent in current_doc_ents]
            fixture_doc_ents_simplified = [[(s.start, s.end) for s in ent.spans] for ent in fixture_doc_ents]

            self.assertEqual(
                current_doc_ents_simplified,
                fixture_doc_ents_simplified,
                msg=f"Current parse produces different SpanGroups for `{annotation_name}`",
            )

            current_doc_ent_boxes = [
                [list(box.xy_coordinates) + [box.page] for box in ent.boxes] for ent in current_doc_ents
            ]
            fixture_doc_ent_boxes = [
                [list(box.xy_coordinates) + [box.page] for box in ent.boxes] for ent in current_doc_ents
            ]

            self.assertAlmostEqual(
                current_doc_ent_boxes,
                fixture_doc_ent_boxes,
                places=3,
                msg=f"Boxes generated for `{annotation_name}` have changed.",
            )

        compare_entities(self.doc.tokens, fixture_doc.tokens, "tokens")
        compare_entities(self.doc.rows, fixture_doc.rows, "rows")
        compare_entities(self.doc.pages, fixture_doc.pages, "pages")



"""


@chrisw, @kylel

"""


from abc import abstractmethod
from typing import List

import numpy as np
from ncls import NCLS

from .annotation import Annotation
from .box import Box
from .entity import Entity


class Indexer:
    """Stores an index for a particular collection of Annotations.
    Indexes in this library focus on *INTERSECT* relations."""

    @abstractmethod
    def find(self, query: Annotation) -> List[Annotation]:
        """Returns all matching Annotations given a suitable query"""
        raise NotImplementedError()


class EntitySpanIndexer(Indexer):
    """
    Manages a data structure for locating overlapping Entity Spans.
    Builds a static nested containment list from Entity Spans
    and accepts other Entity as search probes.

    See: https://github.com/biocore-ntnu/ncls

    [citation]
    Alexander V. Alekseyenko, Christopher J. Lee;
    Nested Containment List (NCList): a new algorithm for accelerating interval query of genome
      alignment and interval databases, Bioinformatics,
    Volume 23, Issue 11, 1 June 2007, Pages 1386–1393, https://doi.org/10.1093/bioinformatics/btl647
    """

    def __init__(self, entities: List[Entity]) -> None:
        starts = []
        ends = []
        ids = []

        for id, entity in enumerate(entities):
            for span in entity.spans:
                starts.append(span.start)
                ends.append(span.end)
                ids.append(id)

        self._entities = entities
        self._index = NCLS(
            np.array(starts, dtype=np.int32), np.array(ends, dtype=np.int32), np.array(ids, dtype=np.int32)
        )

        self._ensure_disjoint()

    def _ensure_disjoint(self) -> None:
        """
        Constituent span groups must be fully disjoint.
        Ensure the integrity of the built index.
        """
        for entity in self._entities:
            for span in entity.spans:
                matches = [match for match in self._index.find_overlap(span.start, span.end)]
                match_ids = [
                    matched_id for _start, _end, matched_id in self._index.find_overlap(span.start, span.end)
                ]
                if len(match_ids) > 1:
                    matches = [self._entities[match_id].to_json() for match_id in match_ids]
                    raise ValueError(
                        f"Detected overlap! While processing the Span {span} as part of query Entity {entity.to_json()}, we found that it overlaps with existing Entity(s):\n"
                        + "\n".join([f"\t{i}\t{m} " for i, m in zip(match_ids, matches)])
                    )

    def find(self, query: Entity) -> List[Entity]:
        if not isinstance(query, Entity):
            raise TypeError(f"EntityIndexer only works with `query` that is Entity type")

        if not query.spans:
            return []

        matched_ids = set()

        for span in query.spans:
            for _start, _end, matched_id in self._index.find_overlap(span.start, span.end):
                matched_ids.add(matched_id)

        matched_entities = [self._entities[matched_id] for matched_id in matched_ids]

        # TODO: kylel - is this necessary? seems like already does this (see tests)
        # Retrieval above doesn't preserve document order; sort here
        # TODO: provide option to return matched span groups in same order as self._entities
        #   (the span groups the index was built with originally)
        return sorted(list(matched_entities))


class EntityBoxIndexer(Indexer):
    """
    Manages a data structure for locating overlapping BoxGroups.
    Builds a static nested containment list from BoxGroups
    and accepts other BoxGroups as search probes.

    @kylel
    """

    def __init__(self, entities: List[Entity], allow_overlap: bool = True) -> None:
        self._entities = entities

        self._box_id_to_entity_id = {}
        self._boxes = []
        box_id = 0
        for i, e in enumerate(entities):
            for box in e.boxes:
                self._boxes.append(box)
                self._box_id_to_entity_id[box_id] = i
                box_id += 1

        self._np_boxes_x1 = np.array([b.l for b in self._boxes])
        self._np_boxes_y1 = np.array([b.t for b in self._boxes])
        self._np_boxes_x2 = np.array([b.l + b.w for b in self._boxes])
        self._np_boxes_y2 = np.array([b.t + b.h for b in self._boxes])
        self._np_boxes_page = np.array([b.page for b in self._boxes])

        if not allow_overlap:
            self._ensure_disjoint()

    def _find_overlap_boxes(self, query: Box) -> List[int]:
        x1, y1, x2, y2 = query.xy_coordinates
        mask = (
            (self._np_boxes_x1 <= x2)
            & (self._np_boxes_x2 >= x1)
            & (self._np_boxes_y1 <= y2)
            & (self._np_boxes_y2 >= y1)
            & (self._np_boxes_page == query.page)
        )
        return np.where(mask)[0].tolist()

    def _find_overlap_entities(self, query: Box) -> List[int]:
        return [self._box_id_to_entity_id[box_id] for box_id in self._find_overlap_boxes(query)]

    def _ensure_disjoint(self) -> None:
        """
        Constituent box groups must be fully disjoint.
        Ensure the integrity of the built index.
        """
        for entity in self._entities:
            for box in entity.boxes:
                match_ids = self._find_overlap_entities(query=box)
                if len(match_ids) > 1:
                    matches = [self._entities[match_id].to_json() for match_id in match_ids]
                    raise ValueError(
                        f"Detected overlap! While processing the Box {box} as part of query Entity {entity.to_json()}, we found that it overlaps with existing Entity(s):\n"
                        + "\n".join([f"\t{i}\t{m} " for i, m in zip(match_ids, matches)])
                    )

    def find(self, query: Entity) -> List[Entity]:
        if not isinstance(query, Entity):
            raise TypeError(f"EntityBoxIndexer only works with `query` that is Entity type")

        if not query.boxes:
            return []

        match_ids = []
        for box in query.boxes:
            match_ids.extend(self._find_overlap_entities(query=box))

        return [self._entities[match_id] for match_id in sorted(set(match_ids))]



"""

@kylel

"""

import json
import os
import pathlib
import re
import unittest

import numpy as np

from papermage.magelib import Box, Document, Entity, Span
from papermage.parsers import PDFPlumberParser
from papermage.rasterizers import PDF2ImageRasterizer


class TestVisualizer(unittest.TestCase):
    def setUp(self) -> None:
        self.fixture_path = pathlib.Path(__file__).parent.parent / "fixtures"
        with open(self.fixture_path / "2304.02623v1.json") as f_in:
            doc_dict = json.load(f_in)
            self.doc = Document.from_json(doc_dict)

        rasterizer = PDF2ImageRasterizer()
        images = rasterizer.rasterize(input_pdf_path=str(self.fixture_path / "2304.02623v1.pdf"), dpi=72)
        rasterizer.attach_images(images=images, doc=self.doc)
        page = self.doc.pages[0]
        page_image = page.images[0]
        entities_on_page = page.tokens



import importlib.metadata
from pathlib import Path


def get_version() -> str:
    """Get the version of the package."""
    # This is a workaround for the fact that if the package is installed
    # in editable mode, the version is not reliability available.
    # Therefore, we check for the existence of a file called EDITABLE,
    # which is not included in the package at distribution time.
    path = Path(__file__).parent / "EDITABLE"
    if path.exists():
        return "dev"

    try:
        # package has been installed, so it has a version number
        # from pyproject.toml
        version = importlib.metadata.version(get_name())
    except importlib.metadata.PackageNotFoundError:
        # package hasn't been installed, so set version to "dev"
        version = "dev"

    return version


def get_name() -> str:
    """Get the name of the package."""
    return "smashed"


def get_name_and_version() -> str:
    """Get the name and version of the package."""
    return f"{get_name()}=={get_version()}"



from typing import Union
from charset_normalizer import detect


def maybe_normalize(text: Union[str, bytes]) -> str:
    """If text is bytes, detect encoding and decode to str."""

    if isinstance(text, bytes):
        enc = str(detect(text)['encoding'])
        text = text.decode(enc)

    return text



"""
Tests for SVM Word Predictor

@kylel
"""

import json
import os
import unittest
from typing import List, Optional, Set

import numpy as np

from papermage.magelib import Document, Entity, Span
from papermage.predictors import SVMWordPredictor
from papermage.predictors.sklearn_predictors.word_predictor import SVMClassifier


class TestSVMClassifier(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.fixture_path = os.path.join(os.path.dirname(__file__), "../fixtures/")
        self.classifier = SVMClassifier.from_path(
            tar_path=os.path.join(self.fixture_path, "svm_word_predictor/svm_word_predictor.tar.gz")
        )
        with open(os.path.join(self.fixture_path, "svm_word_predictor/pos_words.txt")) as f_in:
            self.pos_words = [line.strip() for line in f_in]
        with open(os.path.join(self.fixture_path, "svm_word_predictor/neg_words.txt")) as f_in:
            self.neg_words = [line.strip() for line in f_in]

    def test_batch_predict_unit(self):
        pos_words = [
            "wizard-of-oz",
            "moment-to-moment",
            "batch-to-batch",
            "Seven-day-old",
            "slow-to-fast",
            "HTLV-1-associated",
            "anti-E-selectin",
        ]
        neg_words = [
            "sig-nal-to-noise",
            "nonre-turn-to-zero",
            "comput-er-assisted",
            "concentra-tion-dependent",
            "ob-ject-oriented",
            "cog-nitive-behavioral",
            "deci-sion-makers",
        ]
        THRESHOLD = -1.5
        pos_results = self.classifier.batch_predict(words=pos_words, threshold=THRESHOLD)
        self.assertEqual(len(pos_results), len(pos_words))
        self.assertTrue(all([r.is_edit is False for r in pos_results]))
        neg_results = self.classifier.batch_predict(words=neg_words, threshold=THRESHOLD)
        self.assertEqual(len(neg_results), len(neg_words))
        self.assertTrue(all([r.is_edit is True for r in neg_results]))

    def test_batch_predict_eval(self):
        """
        As a guideline, we want Recall to be close to 1.0 because we want
        the model to favor predicting things as "negative" (i.e. not an edit).
        If the classifier predicts a "1", then essentially we don't do anything.
        Meaning in all cases where the ground truth is "1" (dont do anything),
        we want to recover all these cases nearly perfectly, and ONLY
        take action when absolutely safe.

        THRESHOLD = -1.7    --> P: 0.9621262458471761 R: 1.0
        THRESHOLD = -1.6    --> P: 0.9674346429879954 R: 1.0
        THRESHOLD = -1.5    --> P: 0.9716437941036409 R: 1.0
        THRESHOLD = -1.4    --> P: 0.9755705281460552 R: 0.9999554446622705
        THRESHOLD = -1.0    --> P: 0.9866772193641999 R: 0.9998217786490822
        THRESHOLD = -0.5    --> P: 0.9955352184633155 R: 0.9984405631794689
        THRESHOLD = 0.0     --> P: 0.9985657299090135 R: 0.9926483692746391
        THRESHOLD = 1.0     --> P: 0.9997759019944723 R: 0.8944929602566387
        """
        THRESHOLD = -1.5
        preds_pos = self.classifier.batch_predict(words=self.pos_words, threshold=THRESHOLD)
        self.assertEqual(len(preds_pos), len(self.pos_words))
        preds_pos_as_ints = [int(r.is_edit is False) for r in preds_pos]
        tp = sum(preds_pos_as_ints)
        fn = len(preds_pos_as_ints) - tp

        preds_neg = self.classifier.batch_predict(words=self.neg_words, threshold=THRESHOLD)
        self.assertEqual(len(preds_neg), len(self.neg_words))
        preds_neg_as_ints = [int(r.is_edit is True) for r in preds_neg]
        tn = sum(preds_neg_as_ints)
        fp = len(preds_neg_as_ints) - tn

        self.assertEqual(tp + fn + tn + fp, len(preds_pos) + len(preds_neg))

        p = tp / (tp + fp)
        r = tp / (tp + fn)

        # uncomment for debugging
        # print(f"P: {p} R: {r}")

        self.assertGreaterEqual(p, 0.9)
        self.assertGreaterEqual(r, 0.9)

    def test_get_features(self):
        (
            all_features,
            word_id_to_feature_ids,
        ) = self.classifier._get_features(words=self.pos_words)
        self.assertEqual(len(word_id_to_feature_ids), len(self.pos_words))
        self.assertEqual(
            all_features.shape[0],
            sum([len(feature) for feature in word_id_to_feature_ids.values()]),
        )
        (
            all_features,
            word_id_to_feature_ids,
        ) = self.classifier._get_features(words=self.neg_words)
        self.assertEqual(len(word_id_to_feature_ids), len(self.neg_words))
        self.assertEqual(
            all_features.shape[0],
            sum([len(feature) for feature in word_id_to_feature_ids.values()]),
        )

    def test_exception_with_start_or_end_hyphen(self):
        words = ["-wizard-of-", "wizard-of-"]
        for word in words:
            with self.assertRaises(ValueError):
                self.classifier.batch_predict(words=[word], threshold=0.0)



"""

@kylel

"""

import os
import pathlib
import unittest

from papermage.magelib import Document, Entity, Image
from papermage.recipes import CoreRecipe
from tests.test_recipes.core_recipe_fixtures import (  # SEGMENT_OF_WORD_JSONS,
    BASE64_PAGE_IMAGE,
    FIRST_3_BLOCKS_JSON,
    FIRST_5_ROWS_JSON,
    FIRST_10_TOKENS_JSON,
    FIRST_10_VILA_JSONS,
    FIRST_1000_SYMBOLS,
    PAGE_JSON,
)


def round_all_floats(d: dict):
    import numbers

    def formatfloat(x):
        return "%.4g" % float(x)

    def pformat(dictionary, function):
        if isinstance(dictionary, dict):
            return {key: pformat(value, function) for key, value in dictionary.items()}
        if isinstance(dictionary, list):
            return [pformat(element, function) for element in dictionary]
        if isinstance(dictionary, numbers.Number):
            return function(dictionary)
        return dictionary

    return pformat(d, formatfloat)


class TestCoreRecipe(unittest.TestCase):
    def setUp(self):
        self.fixture_path = pathlib.Path(__file__).parent.parent / "fixtures"
        self.recipe = CoreRecipe()

    def test_stability(self):
        self.recipe.from_path(pdfpath=self.fixture_path / "1903.10676.pdf")
        self.recipe.from_path(pdfpath=self.fixture_path / "2304.02623v1.pdf")
        self.recipe.from_path(pdfpath=self.fixture_path / "2020.acl-main.447.pdf")
        self.recipe.from_path(pdfpath=self.fixture_path / "4be952924cd565488b4a239dc6549095029ee578.pdf")
        self.recipe.from_path(pdfpath=self.fixture_path / "2023.eacl-main.121.pdf")



"""

@kylel

"""

import unittest

from papermage.magelib import Box


class TestBox(unittest.TestCase):
    def test_create(self):
        with self.assertRaises(AssertionError):
            Box(l=0.2, t=0.09, w=-0.095, h=0.017, page=3)
        with self.assertRaises(AssertionError):
            Box(l=0.2, t=0.09, w=0.095, h=-0.017, page=3)

    def test_to_from_json(self):
        box = Box(l=0.2, t=0.09, w=0.095, h=0.017, page=3)
        self.assertEqual(box.to_json(), [0.2, 0.09, 0.095, 0.017, 3])

        box2 = Box.from_json(box.to_json())
        self.assertEqual(box2.l, 0.2)
        self.assertEqual(box2.t, 0.09)
        self.assertEqual(box2.w, 0.095)
        self.assertEqual(box2.h, 0.017)
        self.assertEqual(box2.page, 3)
        self.assertListEqual(box2.to_json(), [0.2, 0.09, 0.095, 0.017, 3])

    def test_numeric_type_casting(self):
        self.assertListEqual(Box(1, 2, 3, 4, 5.0).to_json(), [1.0, 2.0, 3.0, 4.0, 5])

    def test_overlap(self):
        # with itself
        b1 = Box(0.2, 0.09, 0.095, 0.017, 3)
        self.assertTrue(b1.is_overlap(other=b1))
        # page matters
        b2 = Box(0.2, 0.09, 0.095, 0.017, 2)
        self.assertFalse(b1.is_overlap(other=b2))
        # x-axis shift to right too much
        b3 = Box(0.3, 0.09, 0.095, 0.017, 3)
        self.assertFalse(b1.is_overlap(other=b3))
        # x-axis shift just a little
        b4 = Box(0.295, 0.09, 0.095, 0.017, 3)
        self.assertTrue(b1.is_overlap(other=b4))
        # y-axis shift to down too much
        b5 = Box(0.2, 0.108, 0.095, 0.017, 3)
        self.assertFalse(b1.is_overlap(other=b5))
        # y-axis shift just a little
        b6 = Box(0.2, 0.106999, 0.095, 0.017, 3)
        self.assertTrue(b1.is_overlap(other=b6))

    def test_xy_coordinates(self):
        box = Box.from_xy_coordinates(x1=0.2, y1=0.09, x2=0.295, y2=0.107, page=3)
        # allow for some floating point errors
        self.assertAlmostEqual(box.l, 0.2)
        self.assertAlmostEqual(box.t, 0.09)
        self.assertAlmostEqual(box.w, 0.095)
        self.assertAlmostEqual(box.h, 0.017)
        self.assertEqual(box.page, 3)
        # it's important that the return value of this does *NOT* look like the JSON serialization
        # hence why we make it a tuple & drop the page.
        # otherwise, it can be confusing if come back to this data at a later date.
        self.assertEqual(box.xy_coordinates, (0.2, 0.09, 0.295, 0.107))

        box = Box.from_xy_coordinates(x1=0.2, y1=0.09, x2=10.295, y2=30.107, page=3, page_width=10, page_height=30)
        # allow for some floating point errors
        self.assertAlmostEqual(box.l, 0.2)
        self.assertAlmostEqual(box.t, 0.09)
        self.assertAlmostEqual(box.w, 9.8)
        self.assertAlmostEqual(box.h, 29.91)
        self.assertEqual(box.page, 3)
        # it's important that the return value of this does *NOT* look like the JSON serialization
        # hence why we make it a tuple & drop the page.
        # otherwise, it can be confusing if come back to this data at a later date.
        self.assertEqual(box.xy_coordinates, (0.2, 0.09, 10, 30))

    def test_relative_absolute(self):
        box = Box(l=0.2, t=0.09, w=0.095, h=0.017, page=3)
        box_abs = box.to_absolute(page_width=100, page_height=300)
        self.assertAlmostEqual(box_abs.l, 20.0)
        self.assertAlmostEqual(box_abs.t, 27.0)
        self.assertAlmostEqual(box_abs.w, 9.5)
        self.assertAlmostEqual(box_abs.h, 5.1)
        self.assertEqual(box_abs.page, 3)
        box_rel = box_abs.to_relative(page_width=100, page_height=300)
        self.assertAlmostEqual(box_rel.l, 0.2)
        self.assertAlmostEqual(box_rel.t, 0.09)
        self.assertAlmostEqual(box_rel.w, 0.095)
        self.assertAlmostEqual(box_rel.h, 0.017)
        self.assertEqual(box_rel.page, 3)

    def test_center(self):
        box = Box(l=0.2, t=0.09, w=0.095, h=0.017, page=3)
        x, y = box.center
        self.assertAlmostEqual(x, 0.2475)
        self.assertAlmostEqual(y, 0.0985)

    def test_create_enclosing_box(self):
        # nonempty
        with self.assertRaises(ValueError):
            Box.create_enclosing_box(boxes=[])
        # singleton
        box = Box.create_enclosing_box(boxes=[Box(l=0.2, t=0.09, w=0.095, h=0.017, page=3)])
        self.assertAlmostEqual(box.l, 0.2)
        self.assertAlmostEqual(box.t, 0.09)
        self.assertAlmostEqual(box.w, 0.095)
        self.assertAlmostEqual(box.h, 0.017)
        self.assertEqual(box.page, 3)
        # proper behavior
        box2 = Box(0.15, 0.05, 0.01, 0.01, 3)
        box3 = Box(0.05, 0.12, 0.01, 0.01, 3)
        box_new = Box.create_enclosing_box(boxes=[box, box2, box3])
        self.assertAlmostEqual(box_new.l, 0.05)
        self.assertAlmostEqual(box_new.t, 0.05)
        self.assertAlmostEqual(box_new.w, 0.245)
        self.assertAlmostEqual(box_new.h, 0.08)
        self.assertEqual(box_new.page, 3)



"""

@kylel

"""

import unittest

from papermage.magelib import (
    Document,
    EntitiesFieldName,
    Entity,
    MetadataFieldName,
    RelationsFieldName,
    SymbolsFieldName,
)


class TestDocument(unittest.TestCase):
    def test_annotate(self):
        doc = Document("This is a test document!")
        tokens = [
            Entity.from_json({"spans": [[0, 4]]}),
            Entity.from_json({"spans": [[5, 7]]}),
            Entity.from_json({"spans": [[8, 9]]}),
            Entity.from_json({"spans": [[10, 14]]}),
            Entity.from_json({"spans": [[15, 23]]}),
            Entity.from_json({"spans": [[23, 24]]}),
        ]
        with self.assertRaises(AttributeError) as e:
            doc.tokens
        # annotate
        doc.annotate_entity(field_name="tokens", entities=tokens)
        self.assertEqual(len(doc.tokens), len(tokens))
        for t1, t2 in zip(doc.tokens, tokens):
            self.assertEqual(t1, t2)
        # get
        self.assertEqual(doc.tokens, doc.get_entity(field_name="tokens"))
        # remove
        doc.remove_entity(field_name="tokens")
        with self.assertRaises(AttributeError) as e:
            doc.tokens

    def test_empty_annotations_work(self):
        doc = Document("This is a test document!")
        doc.annotate_entity(field_name="my_cool_field", entities=[])
        self.assertEqual(doc.my_cool_field, [])

    def test_metadata_serializes(self):
        symbols = "Hey there y'all!"
        doc = Document(symbols=symbols)
        doc.metadata["a"] = {"b": "c"}
        self.assertDictEqual(
            doc.to_json(),
            {
                SymbolsFieldName: symbols,
                MetadataFieldName: {"a": {"b": "c"}},
                EntitiesFieldName: {},
                RelationsFieldName: {},
            },
        )

    def test_metadata_deserializes(self):
        metadata = {"a": {"b": "c"}}
        symbols = "Hey again peeps!"
        input_json = {
            SymbolsFieldName: symbols,
            MetadataFieldName: metadata,
            EntitiesFieldName: {},
            RelationsFieldName: {},
        }

        doc = Document.from_json(input_json)

        self.assertEqual(symbols, doc.symbols)
        self.assertDictEqual(metadata, doc.metadata.to_json())

    def test_metadata_deserializes_when_empty(self):
        symbols = "That's all folks!"
        input_json = {
            SymbolsFieldName: symbols,
            MetadataFieldName: {},
            EntitiesFieldName: {},
            RelationsFieldName: {},
        }

        doc = Document.from_json(input_json)

        self.assertEqual(symbols, doc.symbols)
        self.assertEqual(0, len(doc.metadata))

    def test_cross_referencing(self):
        doc = Document("This is a test document!")
        # boxes are in a top-left to bottom-right diagonal fashion (same page)
        tokens = [
            Entity.from_json({"spans": [[0, 4]], "boxes": [[0, 0, 0.5, 0.5, 0]]}),
            Entity.from_json({"spans": [[5, 7]], "boxes": [[1, 1, 0.5, 0.5, 0]]}),
            Entity.from_json({"spans": [[8, 9]], "boxes": [[2, 2, 0.5, 0.5, 0]]}),
            Entity.from_json({"spans": [[10, 14]], "boxes": [[3, 3, 0.5, 0.5, 0]]}),
            Entity.from_json({"spans": [[15, 23]], "boxes": [[4, 4, 0.5, 0.5, 0]]}),
            Entity.from_json({"spans": [[23, 24]], "boxes": [[5, 5, 0.5, 0.5, 0]]}),
        ]
        # boxes are also same diagonal fashion, but bigger.
        # last box super big on wrong page.
        chunks = [
            Entity.from_json({"spans": [[0, 9]], "boxes": [[0, 0, 2.01, 2.01, 0]]}),
            Entity.from_json({"spans": [[12, 23]], "boxes": [[3.0, 3.0, 4.0, 4.0, 0]]}),
            Entity.from_json({"spans": [[23, 24]], "boxes": [[0, 0, 10.0, 10.0, 1]]}),
        ]
        doc.annotate_entity(field_name="tokens", entities=tokens)
        doc.annotate_entity(field_name="chunks", entities=chunks)

        # find by span is the default overload of Entity.__attr__
        self.assertListEqual(doc.chunks[0].tokens, tokens[0:3])
        self.assertListEqual(doc.chunks[1].tokens, tokens[3:5])
        self.assertListEqual(doc.chunks[2].tokens, [tokens[5]])

        # backwards
        self.assertListEqual(doc.tokens[0].chunks, [chunks[0]])
        self.assertListEqual(doc.tokens[1].chunks, [chunks[0]])
        self.assertListEqual(doc.tokens[2].chunks, [chunks[0]])
        self.assertListEqual(doc.tokens[3].chunks, [chunks[1]])
        self.assertListEqual(doc.tokens[4].chunks, [chunks[1]])
        self.assertListEqual(doc.tokens[5].chunks, [chunks[2]])

        # find by span works fine
        self.assertListEqual(doc.chunks[0].tokens, doc.find_by_span(query=doc.chunks[0], field_name="tokens"))
        self.assertListEqual(doc.chunks[1].tokens, doc.find_by_span(query=doc.chunks[1], field_name="tokens"))
        self.assertListEqual(doc.chunks[2].tokens, doc.find_by_span(query=doc.chunks[2], field_name="tokens"))

        # backwards
        self.assertListEqual(doc.tokens[0].chunks, doc.find_by_span(query=doc.tokens[0], field_name="chunks"))
        self.assertListEqual(doc.tokens[1].chunks, doc.find_by_span(query=doc.tokens[1], field_name="chunks"))
        self.assertListEqual(doc.tokens[2].chunks, doc.find_by_span(query=doc.tokens[2], field_name="chunks"))
        self.assertListEqual(doc.tokens[3].chunks, doc.find_by_span(query=doc.tokens[3], field_name="chunks"))
        self.assertListEqual(doc.tokens[4].chunks, doc.find_by_span(query=doc.tokens[4], field_name="chunks"))
        self.assertListEqual(doc.tokens[5].chunks, doc.find_by_span(query=doc.tokens[5], field_name="chunks"))

        # find by box
        self.assertListEqual(doc.find_by_box(query=doc.chunks[0], field_name="tokens"), doc.tokens[0:3])
        self.assertListEqual(doc.find_by_box(query=doc.chunks[1], field_name="tokens"), doc.tokens[3:6])
        self.assertListEqual(doc.find_by_box(query=doc.chunks[2], field_name="tokens"), [])

        # backwards
        self.assertListEqual(doc.find_by_box(query=doc.tokens[0], field_name="chunks"), [chunks[0]])
        self.assertListEqual(doc.find_by_box(query=doc.tokens[1], field_name="chunks"), [chunks[0]])
        self.assertListEqual(doc.find_by_box(query=doc.tokens[2], field_name="chunks"), [chunks[0]])
        self.assertListEqual(doc.find_by_box(query=doc.tokens[3], field_name="chunks"), [chunks[1]])
        self.assertListEqual(doc.find_by_box(query=doc.tokens[4], field_name="chunks"), [chunks[1]])
        self.assertListEqual(doc.find_by_box(query=doc.tokens[5], field_name="chunks"), [chunks[1]])

    def test_cross_referencing_with_missing_entity_fields(self):
        """What happens when annotate a Doc with entiites missing spans or boxes?
        How does the cross-referencing operation behave?"""
        # repeat the above test, but with missing spans and boxes
        doc = Document("This is a test document!")
        tokens = [
            Entity.from_json({"spans": [[0, 4]]}),
            Entity.from_json({"spans": [[5, 7]]}),
            Entity.from_json({"spans": [[8, 9]]}),
            Entity.from_json({"spans": [[10, 14]]}),
            Entity.from_json({"spans": [[15, 23]]}),
            Entity.from_json({"spans": [[23, 24]]}),
        ]
        chunks = [
            Entity.from_json({"boxes": [[0, 0, 2.01, 2.01, 0]]}),
            Entity.from_json({"boxes": [[3.0, 3.0, 4.0, 4.0, 0]]}),
            Entity.from_json({"boxes": [[0, 0, 10.0, 10.0, 1]]}),
        ]
        doc.annotate_entity(field_name="tokens", entities=tokens)
        doc.annotate_entity(field_name="chunks", entities=chunks)
        self.assertListEqual(doc.find_by_box(query=doc.chunks[0], field_name="tokens"), [])
        self.assertListEqual(doc.find_by_box(query=doc.chunks[1], field_name="tokens"), [])
        self.assertListEqual(doc.find_by_box(query=doc.chunks[2], field_name="tokens"), [])
        self.assertListEqual(doc.find_by_span(query=doc.chunks[0], field_name="tokens"), [])
        self.assertListEqual(doc.find_by_span(query=doc.chunks[1], field_name="tokens"), [])
        self.assertListEqual(doc.find_by_span(query=doc.chunks[2], field_name="tokens"), [])
        self.assertListEqual(doc.find_by_box(query=doc.tokens[0], field_name="chunks"), [])
        self.assertListEqual(doc.find_by_box(query=doc.tokens[1], field_name="chunks"), [])
        self.assertListEqual(doc.find_by_box(query=doc.tokens[2], field_name="chunks"), [])
        self.assertListEqual(doc.find_by_span(query=doc.tokens[0], field_name="chunks"), [])
        self.assertListEqual(doc.find_by_span(query=doc.tokens[1], field_name="chunks"), [])
        self.assertListEqual(doc.find_by_span(query=doc.tokens[2], field_name="chunks"), [])

    def test_query(self):
        doc = Document("This is a test document!")
        tokens = [
            Entity.from_json({"spans": [[0, 4]], "boxes": [[0, 0, 0.5, 0.5, 0]]}),
            Entity.from_json({"spans": [[5, 7]], "boxes": [[1, 1, 0.5, 0.5, 0]]}),
            Entity.from_json({"spans": [[8, 9]], "boxes": [[2, 2, 0.5, 0.5, 0]]}),
            Entity.from_json({"spans": [[10, 14]], "boxes": [[3, 3, 0.5, 0.5, 0]]}),
            Entity.from_json({"spans": [[15, 23]], "boxes": [[4, 4, 0.5, 0.5, 0]]}),
            Entity.from_json({"spans": [[23, 24]], "boxes": [[5, 5, 0.5, 0.5, 0]]}),
        ]
        chunks = [
            Entity.from_json({"spans": [[0, 9]], "boxes": [[0, 0, 2.01, 2.01, 0]]}),
            Entity.from_json({"spans": [[12, 23]], "boxes": [[3.0, 3.0, 4.0, 4.0, 0]]}),
            Entity.from_json({"spans": [[23, 24]], "boxes": [[0, 0, 10.0, 10.0, 1]]}),
        ]
        doc.annotate_entity(field_name="tokens", entities=tokens)
        doc.annotate_entity(field_name="chunks", entities=chunks)

        # test query by span
        self.assertListEqual(
            doc.find_by_span(query=doc.chunks[0], field_name="tokens"),
            doc.find(query=doc.chunks[0].spans[0], field_name="tokens"),
        )
        # test query by box
        self.assertListEqual(
            doc.find_by_box(query=doc.chunks[0], field_name="tokens"),
            doc.find(query=doc.chunks[0].boxes[0], field_name="tokens"),
        )
        # calling wrong method w input type should fail
        with self.assertRaises(TypeError):
            doc.find_by_box(query=doc.chunks[0].spans[0], field_name="tokens")
        with self.assertRaises(TypeError):
            doc.find_by_span(query=doc.chunks[0].boxes[0], field_name="tokens")
        with self.assertRaises(TypeError):
            doc.find(query=doc.chunks[0], field_name="tokens")



"""

Rectangular region on a document.

@kylel

"""

import logging
from typing import List, Optional, Tuple, Union

import numpy as np

from .span import Span


class _BoxSpan(Span):
    def __init__(self, start: float, end: float):
        self.start = start  # type: ignore
        self.end = end  # type: ignore


class Box:

    __slots__ = ["l", "t", "w", "h", "page"]

    def __init__(self, l: float, t: float, w: float, h: float, page: int):
        assert w >= 0.0, "Box width cant be negative"
        assert h >= 0.0, "Box height cant be negative"
        self.l = float(l)
        self.t = float(t)
        self.w = float(w)
        self.h = float(h)
        self.page = int(page)

    def to_json(self) -> List[Union[float, int]]:
        """Returns whatever representation is JSON compatible"""
        return [self.l, self.t, self.w, self.h, self.page]

    @classmethod
    def from_json(cls, box_json: List[Union[float, int]]) -> "Box":
        """Recreates the object from the JSON serialization"""
        l, t, w, h, page = box_json
        return Box(l=l, t=t, w=w, h=h, page=int(page))

    def __repr__(self):
        return f"Box{self.to_json()}"

    @classmethod
    def from_xy_coordinates(
        cls,
        x1: float,
        y1: float,
        x2: float,
        y2: float,
        page: int,
        page_width: Optional[float] = None,
        page_height: Optional[float] = None,
    ):
        """Create a Box from the top-left and bottom-right coordinates.

        If the page width and height are provided and coordinates lie outside the page, they are clipped to the
        page width and height.
        """

        assert x2 >= x1, "Requires x2 >= x1"
        assert y2 >= y1, "Requires y2 >= y1"
        _x1, _x2, _y1, _y2 = x1, x2, y1, y2
        if page_width is not None:
            _x1, _x2 = np.clip([x1, x2], 0, page_width)
        if page_height is not None:
            _y1, _y2 = np.clip([y1, y2], 0, page_height)

        if (_x1, _y1, _x2, _y2) != (x1, y1, x2, y2):
            logging.warning(
                f"The coordinates ({x1}, {y1}, {x2}, {y2}) are not valid and converted to ({_x1}, {_y1}, {_x2}, {_y2})."
            )

        return Box(l=_x1, t=_y1, w=_x2 - _x1, h=_y2 - _y1, page=page)

    @property
    def xy_coordinates(self) -> Tuple[float, float, float, float]:
        return self.l, self.t, self.l + self.w, self.t + self.h

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Box):
            return False
        return (
            self.l == other.l
            and self.t == other.t
            and self.w == other.w
            and self.h == other.h
            and self.page == other.page
        )

    def to_relative(self, page_width: float, page_height: float) -> "Box":
        """Get the relative coordinates of self based on page_width, page_height."""
        return self.__class__(
            l=float(self.l) / page_width,
            t=float(self.t) / page_height,
            w=float(self.w) / page_width,
            h=float(self.h) / page_height,
            page=self.page,
        )

    def to_absolute(self, page_width: int, page_height: int) -> "Box":
        """Get the absolute coordinates of self based on page_width, page_height."""
        return self.__class__(
            l=self.l * page_width,
            t=self.t * page_height,
            w=self.w * page_width,
            h=self.h * page_height,
            page=self.page,
        )

    @property
    def center(self) -> Tuple[float, float]:
        return self.l + self.w / 2, self.t + self.h / 2

    def is_overlap(self, other: "Box") -> bool:
        """Whether self overlaps with the other Box object"""

        if self.page != other.page:
            return False

        self_x1, self_y1, self_x2, self_y2 = self.xy_coordinates
        other_x1, other_y1, other_x2, other_y2 = other.xy_coordinates

        # check x-axis
        span_x_self = _BoxSpan(start=self_x1, end=self_x2)
        span_x_other = _BoxSpan(start=other_x1, end=other_x2)
        if not span_x_self.is_overlap(span_x_other):
            return False

        # check y-axis
        span_y_self = _BoxSpan(start=self_y1, end=self_y2)
        span_y_other = _BoxSpan(start=other_y1, end=other_y2)
        if not span_y_self.is_overlap(span_y_other):
            return False

        return True

    @classmethod
    def create_enclosing_box(cls, boxes: List["Box"]) -> "Box":
        """Create the narrowest Box that completely encloses all the input Boxes."""
        if not boxes:
            raise ValueError(f"`boxes` should be non-empty.")
        unique_pages = {box.page for box in boxes}
        if len(unique_pages) != 1:
            raise ValueError(f"Boxes not all on same page. Pages={unique_pages}")
        x1 = min([box.l for box in boxes])
        y1 = min([box.t for box in boxes])
        x2 = max([box.l + box.w for box in boxes])
        y2 = max([box.t + box.h for box in boxes])
        return Box(l=x1, t=y1, w=x2 - x1, h=y2 - y1, page=boxes[0].page)



from typing import Any, Dict, List, Optional, Union

import layoutparser as lp
from tqdm import tqdm

from papermage.magelib import (
    Annotation,
    Box,
    Document,
    Entity,
    Image,
    ImagesFieldName,
    Metadata,
    PagesFieldName,
)
from papermage.predictors.base_predictor import BasePredictor


class LPBlockPredictor(BasePredictor):
    REQUIRED_BACKENDS = ["layoutparser"]
    REQUIRED_DOCUMENT_FIELDS = [PagesFieldName, ImagesFieldName]

    def __init__(self, model):
        self.model = model

    @classmethod
    def from_pretrained(
        cls,
        config_path: str = "lp://efficientdet/PubLayNet",
        model_path: Optional[str] = None,
        label_map: Optional[Dict] = None,
        extra_config: Optional[Dict] = None,
        device: Optional[str] = None,
    ):
        """Initialize a pre-trained layout detection model from
        layoutparser. The parameters currently are the same as the
        default layoutparser Detectron2 models
        https://layout-parser.readthedocs.io/en/latest/api_doc/models.html ,
        and will be updated in the future.
        """

        model = lp.AutoLayoutModel(
            config_path=config_path,
            model_path=model_path,
            label_map=label_map,
            extra_config=extra_config,
            device=device,
        )

        return cls(model)

    def postprocess(self, model_outputs: lp.Layout, page_index: int, image: Image) -> List[Annotation]:
        """Convert the model outputs into the papermage format

        Args:
            model_outputs (lp.Layout):
                The layout detection results from layoutparser for
                a page image
            page_index (int):
                The index of the current page, used for creating the
                `Box` object
            image (Image):
                The image of the current page, used for converting
                to relative coordinates for the box objects

        Returns:
            List[Annotation]:
            The detected layout stored in a list of Entities.
        """

        # block.coordinates returns the left, top, bottom, right coordinates

        page_width, page_height = image.pilimage.size

        return [
            Entity(
                boxes=[
                    Box(
                        l=block.coordinates[0],
                        t=block.coordinates[1],
                        w=block.width,
                        h=block.height,
                        page=page_index,
                    ).to_relative(
                        page_width=page_width,
                        page_height=page_height,
                    )
                ],
                metadata=Metadata(type=block.type),
            )
            for block in model_outputs
        ]

    def _predict(self, doc: Document) -> List[Annotation]:
        """Returns a list of Entities for the detected layouts for all pages

        Args:
            document (Document):
                The input document object

        Returns:
            List[Annotation]:
                The returned Entities for the detected layouts for all pages
        """
        document_prediction = []

        images = doc.get_entity(field_name=ImagesFieldName)
        for image_index, image in enumerate(tqdm(images)):
            model_outputs = self.model.detect(image.pilimage)
            document_prediction.extend(self.postprocess(model_outputs, image_index, image))

        return document_prediction



"""

@kylel

"""

import unittest

from papermage.magelib import Box, Entity, Metadata, Span


class TestEntity(unittest.TestCase):
    def test_create_empty(self):
        with self.assertRaises(ValueError):
            e = Entity()
        # both nones not OK
        with self.assertRaises(ValueError):
            e = Entity(spans=None)
        with self.assertRaises(ValueError):
            e = Entity(boxes=None)
        # empty lists also not ok
        with self.assertRaises(ValueError):
            e = Entity(spans=[])
        with self.assertRaises(ValueError):
            e = Entity(boxes=[])
        # metadata-only also not ok
        m = Metadata(x=123, y=456)
        with self.assertRaises(ValueError):
            e = Entity(metadata=m)

    def test_create(self):
        # just spans
        spans = [Span(0, 1), Span(2, 3)]
        entity = Entity(spans=spans)
        self.assertListEqual(entity.spans, spans)
        self.assertListEqual(entity.boxes, [])

        # just boxes
        boxes = [Box(0, 0, 0, 0, 0), Box(1, 1, 1, 1, 1)]
        entity = Entity(boxes=boxes)
        self.assertListEqual(entity.boxes, boxes)
        self.assertListEqual(entity.spans, [])

        # both spans and boxes
        entity = Entity(spans=spans, boxes=boxes)
        self.assertListEqual(entity.spans, spans)
        self.assertListEqual(entity.boxes, boxes)

        # auto-creates metadata
        self.assertDictEqual(entity.metadata.to_json(), {})

        # manual-created metadata
        metadata = Metadata(x=123, y=456)
        entity = Entity(spans=spans, boxes=boxes, metadata=metadata)
        self.assertIs(entity.metadata, metadata)

    def test_to_from_json(self):
        spans = [Span(0, 1), Span(2, 3)]
        boxes = [Box(0, 0, 0, 0, 0), Box(1, 1, 1, 1, 1)]
        metadata = Metadata(x=123, y=456)

        #  spans only
        entity = Entity(spans=spans)
        self.assertEqual(entity.to_json(), {"spans": [[0, 1], [2, 3]]})
        entity2 = Entity.from_json(entity.to_json())
        self.assertDictEqual(entity2.to_json(), entity.to_json())

        #  boxes only
        entity = Entity(boxes=boxes)
        self.assertEqual(entity.to_json(), {"boxes": [[0.0, 0.0, 0.0, 0.0, 0], [1.0, 1.0, 1.0, 1.0, 1]]})
        entity2 = Entity.from_json(entity.to_json())
        self.assertDictEqual(entity2.to_json(), entity.to_json())

        # both
        entity = Entity(spans=spans, boxes=boxes)
        self.assertEqual(
            entity.to_json(),
            {"spans": [[0, 1], [2, 3]], "boxes": [[0.0, 0.0, 0.0, 0.0, 0], [1.0, 1.0, 1.0, 1.0, 1]]},
        )
        entity2 = Entity.from_json(entity.to_json())
        self.assertDictEqual(entity2.to_json(), entity.to_json())

        # incl metadata
        entity = Entity(spans=spans, boxes=boxes, metadata=metadata)
        self.assertEqual(
            entity.to_json(),
            {
                "spans": [[0, 1], [2, 3]],
                "boxes": [[0.0, 0.0, 0.0, 0.0, 0], [1.0, 1.0, 1.0, 1.0, 1]],
                "metadata": {"x": 123, "y": 456},
            },
        )
        entity2 = Entity.from_json(entity.to_json())
        self.assertDictEqual(entity2.to_json(), entity.to_json())



"""

@shannons, @kylel

"""

import json
import os
import pathlib
import unittest

from PIL import Image

from papermage.magelib import Document
from papermage.parsers.pdfplumber_parser import PDFPlumberParser
from papermage.predictors import IVILATokenClassificationPredictor, LPBlockPredictor
from papermage.rasterizers.rasterizer import PDF2ImageRasterizer


class TestFigureVilaPredictors(unittest.TestCase):
    @classmethod
    def setUp(cls):
        cls.fixture_path = pathlib.Path(__file__).parent.parent / "fixtures"

        cls.DOCBANK_LABEL_MAP = {
            "0": "paragraph",
            "1": "title",
            "2": "equation",
            "3": "reference",
            "4": "section",
            "5": "list",
            "6": "table",
            "7": "caption",
            "8": "author",
            "9": "abstract",
            "10": "footer",
            "11": "date",
            "12": "figure",
        }
        cls.DOCBANK_LABEL_MAP = {int(key): val for key, val in cls.DOCBANK_LABEL_MAP.items()}

        cls.S2VL_LABEL_MAP = {
            "0": "Title",
            "1": "Author",
            "2": "Abstract",
            "3": "Keywords",
            "4": "Section",
            "5": "Paragraph",
            "6": "List",
            "7": "Bibliography",
            "8": "Equation",
            "9": "Algorithm",
            "10": "Figure",
            "11": "Table",
            "12": "Caption",
            "13": "Header",
            "14": "Footer",
            "15": "Footnote",
        }

        cls.S2VL_LABEL_MAP = {int(key): val for key, val in cls.S2VL_LABEL_MAP.items()}

    def test_vila_predictors(self):
        layout_predictor = LPBlockPredictor.from_pretrained("lp://efficientdet/PubLayNet")

        pdfplumber_parser = PDFPlumberParser()
        rasterizer = PDF2ImageRasterizer()

        doc = pdfplumber_parser.parse(input_pdf_path=self.fixture_path / "1903.10676.pdf")
        images = rasterizer.rasterize(input_pdf_path=self.fixture_path / "1903.10676.pdf", dpi=72)
        doc.annotate_images(images)

        layout_regions = layout_predictor.predict(doc)
        doc.annotate_entity(field_name="blocks", entities=layout_regions)

        predictor_with_blocks = IVILATokenClassificationPredictor.from_pretrained(
            "allenai/ivila-block-layoutlm-finetuned-docbank"
        )
        results_with_blocks = predictor_with_blocks.predict(doc, subpage_per_run=2)

        predictor_with_rows = IVILATokenClassificationPredictor.from_pretrained(
            "allenai/ivila-row-layoutlm-finetuned-s2vl-v2"
        )
        results_with_rows = predictor_with_rows.predict(doc, subpage_per_run=2)



"""

Protocol for creating token streams from a document

@kylel, shannons, bnewm0609

"""

from abc import abstractmethod
from typing import Protocol

from papermage.magelib import Document


class Parser(Protocol):
    @abstractmethod
    def parse(self, input_pdf_path: str, **kwargs) -> Document:
        """Given an input PDF return a Document with at least symbols

        Args:
            input_pdf_path (str): Path to the input PDF to process

        Returns:
            Document: Depending on parser support at least symbols in the PDF
        """
        raise NotImplementedError



"""

This file rewrites the PDFPredictor classes in
https://github.com/allenai/VILA/blob/dd242d2fcbc5fdcf05013174acadb2dc896a28c3/src/vila/predictors.py#L1
to reduce the dependency on the VILA package.

@shannons, @kylel

"""


import inspect
import itertools
from abc import abstractmethod
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

import torch
from tqdm import tqdm
from vila.predictors import LayoutIndicatorPDFPredictor, SimplePDFPredictor

from papermage.magelib import (
    Annotation,
    BlocksFieldName,
    Document,
    Entity,
    Metadata,
    PagesFieldName,
    RowsFieldName,
    Span,
    TokensFieldName,
)
from papermage.predictors.base_predictor import BasePredictor

# Two constants for the constraining the size of the page for
# inputs to the model.
# TODO: Move this to somewhere else.
MAX_PAGE_WIDTH = 1000
MAX_PAGE_HEIGHT = 1000

# these are the labels that are used in the VILA model
VILA_LABELS = [
    "Title",
    "Author",
    "Abstract",
    "Keywords",
    "Section",
    "Paragraph",
    "List",
    "Bibliography",
    "Equation",
    "Algorithm",
    "Figure",
    "Table",
    "Caption",
    "Header",
    "Footer",
    "Footnote"
]


# util
def columns_used_in_model_inputs(model):
    signature = inspect.signature(model.forward)
    signature_columns = list(signature.parameters.keys())
    return signature_columns


# util
def normalize_bbox(
    bbox,
    page_width,
    page_height,
    target_width,
    target_height,
):
    """
    Normalize bounding box to the target size.
    """

    x1, y1, x2, y2 = bbox

    # Right now only execute this for only "large" PDFs
    # TODO: Change it for all PDFs
    if page_width > target_width or page_height > target_height:
        x1 = float(x1) / page_width * target_width
        x2 = float(x2) / page_width * target_width
        y1 = float(y1) / page_height * target_height
        y2 = float(y2) / page_height * target_height

    return (x1, y1, x2, y2)


# util
def shift_index_sequence_to_zero_start(sequence):
    """
    Shift a sequence to start at 0.
    """
    sequence_start = min(sequence)
    return [i - sequence_start for i in sequence]


# util
def get_visual_group_id(token: Entity, field_name: str, defaults=-1) -> int:
    if not hasattr(token, field_name):
        return defaults
    field_value = getattr(token, field_name)
    if len(field_value) == 0 or field_value[0].id is None:
        return defaults
    return field_value[0].id


# util
def convert_document_page_to_pdf_dict(doc: Document, page_width: int, page_height: int) -> Dict[str, List]:
    """Convert a document to a dictionary of the form:
        {
            'words': ['word1', 'word2', ...],
            'bbox': [[x1, y1, x2, y2], [x1, y1, x2, y2], ...],
            'block_ids': [0, 0, 0, 1 ...],
            'line_ids': [0, 1, 1, 2 ...],
            'labels': [0, 0, 0, 1 ...], # could be empty
        }

    Args:
        document (Document):
            The input document object
        page_width (int):
            Typically the transformer model requires to use
            the absolute coordinates for encoding the coordinates.
            Set the correspnding page_width and page_height to convert the
            relative coordinates to the absolute coordinates.
        page_height (int):
            Typically the transformer model requires to use
            the absolute coordinates for encoding the coordinates.
            Set the correspnding page_width and page_height to convert the
            relative coordinates to the absolute coordinates.

    Returns:
        Dict[str, List]: The pdf_dict object
    """

    token_data = [
        (
            token.symbols_from_spans[0],  # words
            token.boxes[0].to_absolute(page_width=page_width, page_height=page_height).xy_coordinates,  # bbox
            get_visual_group_id(token, RowsFieldName, -1),  # line_ids
            get_visual_group_id(token, BlocksFieldName, -1),  # block_ids
        )
        for token in doc.tokens
    ]

    words, bbox, line_ids, block_ids = (list(l) for l in zip(*token_data))
    line_ids = shift_index_sequence_to_zero_start(line_ids)
    block_ids = shift_index_sequence_to_zero_start(block_ids)

    labels = [None] * len(words)
    # TODO: We provide an empty label list.

    return {
        "words": words,
        "bbox": bbox,
        "block_ids": block_ids,
        "line_ids": line_ids,
        "labels": labels,
    }


# util
def convert_sequence_tagging_to_spans(
    token_prediction_sequence: List,
) -> List[Tuple[int, int, int]]:
    """For a sequence of token predictions, convert them to spans
    of consecutive same predictions.

    Args:
        token_prediction_sequence (List)

    Returns:
        List[Tuple[int, int, int]]: A list of (start, end, label)
            of consecutive prediction of the same label.
    """
    prev_len = 0
    spans = []
    for gp, seq in itertools.groupby(token_prediction_sequence):
        cur_len = len(list(seq))
        spans.append((prev_len, prev_len + cur_len, gp))
        prev_len = prev_len + cur_len
    return spans


class BaseSinglePageTokenClassificationPredictor(BasePredictor):
    REQUIRED_BACKENDS = ["transformers", "torch", "vila"]
    REQUIRED_DOCUMENT_FIELDS = [PagesFieldName, TokensFieldName]
    DEFAULT_SUBPAGE_PER_RUN = 2  # TODO: Might remove this in the future for longformer-like models

    @property
    @abstractmethod
    def VILA_MODEL_CLASS(self):
        pass

    def __init__(self, predictor, subpage_per_run: Optional[int] = None):
        self.predictor = predictor

        # TODO: Make this more robust
        self.id2label = self.predictor.model.config.id2label
        self.label2id = self.predictor.model.config.label2id

        self.subpage_per_run = subpage_per_run or self.DEFAULT_SUBPAGE_PER_RUN

    @classmethod
    def from_pretrained(
        cls,
        model_name_or_path: str,
        preprocessor=None,
        device: Optional[str] = None,
        subpage_per_run: Optional[int] = None,
        **preprocessor_config,
    ):
        predictor = cls.VILA_MODEL_CLASS.from_pretrained(
            model_path=model_name_or_path, preprocessor=preprocessor, device=device, **preprocessor_config
        )

        return cls(predictor, subpage_per_run)

    def predict(self, doc: Document, subpage_per_run: Optional[int] = None) -> List[Annotation]:
        page_prediction_results = []
        for page_id, page in enumerate(doc.pages):
            if page.tokens:
                page_width, page_height = doc.images[page_id].pilimage.size

                pdf_dict = self.preprocess(page, page_width=page_width, page_height=page_height)

                model_predictions = self.predictor.predict(
                    page_data=pdf_dict,
                    page_size=(page_width, page_height),
                    batch_size=subpage_per_run or self.subpage_per_run,
                    return_type="list",
                )

                assert len(model_predictions) == len(
                    page.tokens
                ), f"Model predictions and tokens are not the same length ({len(model_predictions)} != {len(page.tokens)}) for page {page_id}"

                page_prediction_results.extend(self.postprocess(page, model_predictions))

        return page_prediction_results

    def preprocess(self, page: Document, page_width: float, page_height: float) -> Dict:
        # In the latest vila implementations (after 0.4.0), the predictor will
        # handle all other preprocessing steps given the pdf_dict input format.

        return convert_document_page_to_pdf_dict(page, page_width=page_width, page_height=page_height)

    def postprocess(self, doc: Document, model_predictions) -> List[Entity]:
        token_prediction_spans = convert_sequence_tagging_to_spans(model_predictions)

        prediction_spans = []
        for token_start, token_end, label in token_prediction_spans:
            cur_spans = doc.tokens[token_start:token_end]

            start = min([ele.start for ele in cur_spans])
            end = max([ele.end for ele in cur_spans])
            sg = Entity(spans=[Span(start, end)], metadata=Metadata(label=label))
            prediction_spans.append(sg)
        return prediction_spans


class SinglePageTokenClassificationPredictor(BaseSinglePageTokenClassificationPredictor):
    VILA_MODEL_CLASS = SimplePDFPredictor


class IVILATokenClassificationPredictor(BaseSinglePageTokenClassificationPredictor):
    VILA_MODEL_CLASS = LayoutIndicatorPDFPredictor

    @property
    def REQUIRED_DOCUMENT_FIELDS(self) -> List:
        base_reqs = [PagesFieldName, TokensFieldName]
        if self.predictor.preprocessor.config.agg_level == "row":
            base_reqs.append(RowsFieldName)
        elif self.predictor.preprocessor.config.agg_level == "block":
            base_reqs.append(BlocksFieldName)
        return base_reqs



"""

@kylel

"""

import os
import unittest

import numpy as np

from papermage.magelib import Image


class TestImage(unittest.TestCase):
    def test_to_from_array(self):
        imarray = np.array([[[1, 2, 3], [4, 5, 6], [7, 8, 9]], [[11, 12, 13], [14, 15, 16], [17, 18, 19]]])
        image = Image.from_array(imarray=imarray)
        imarray2 = image.to_array()
        self.assertListEqual(imarray.tolist(), imarray2.tolist())

    def test_to_from_array_dimensions(self):
        # wrong number of color channel
        with self.assertRaises(ValueError):
            imarray = np.array([[[1, 2], [4, 5], [7, 8]], [[11, 12], [14, 15], [17, 18]]])
            image = Image.from_array(imarray=imarray)
        # missing some dimension
        with self.assertRaises(ValueError):
            imarray = np.array([[1, 2, 4, 5, 7, 8], [11, 12, 14, 15, 17, 18]])
            image = Image.from_array(imarray=imarray)

    def test_eq(self):
        imarray = np.array([[[1, 2, 3], [4, 5, 6], [7, 8, 9]], [[11, 12, 13], [14, 15, 16], [17, 18, 19]]])
        image1 = Image.from_array(imarray=imarray)
        image2 = Image.from_array(imarray=imarray)
        self.assertEqual(image1, image2)

        imarray2 = np.array([[[1, 2, 3], [4, 5, 6], [7, 8, 9]], [[11, 12, 13], [14, 15, 16], [17, 18, 20]]])
        image3 = Image.from_array(imarray=imarray2)
        self.assertNotEqual(image1, image3)

    def test_greyscale(self):
        imarray = np.array([[[1, 2, 3], [4, 5, 6], [7, 8, 9]], [[11, 12, 13], [14, 15, 16], [17, 18, 19]]])
        image = Image.from_array(imarray=imarray)
        image_grey = image.convert_to_greyscale()
        self.assertListEqual(
            image_grey.to_array().tolist(),
            [[[2, 2, 2], [5, 5, 5], [8, 8, 8]], [[12, 12, 12], [15, 15, 15], [18, 18, 18]]],
        )
        self.assertNotEqual(image, image_grey)
        self.assertEqual(image_grey.mode, "RGB")

    def test_save_open(self):
        # open
        imagefile = os.path.join(os.path.dirname(__file__), "../fixtures/white_page.png")
        image = Image.open(imagefile)
        # save
        imagefile = imagefile.replace("white_page.png", "temp_123_white_page.png")
        image.save(imagefile)
        # clean
        os.remove(imagefile)

    def test_create_rgb_all_white(self):
        image = Image.create_rgb_all_white(width=600, height=800)
        imagefile = os.path.join(os.path.dirname(__file__), "../fixtures/white_page.png")
        image2 = Image.open(imagefile)
        self.assertEqual(image, image2)



import json
import pathlib
import unittest

from papermage.magelib import Entity
from papermage.parsers import PDFPlumberParser
from papermage.predictors.lp_predictors.block_predictor import LPBlockPredictor
from papermage.rasterizers import PDF2ImageRasterizer


class TestLayoutParserPredictor(unittest.TestCase):
    def setUp(self):
        self.fixture_path = pathlib.Path(__file__).parent.parent / "fixtures"
        self.parser = PDFPlumberParser()
        self.rasterizer = PDF2ImageRasterizer()
        self.layout_predictor = LPBlockPredictor.from_pretrained("lp://efficientdet/PubLayNet")

    def test_predict(self):
        input_pdf_path = self.fixture_path / "4be952924cd565488b4a239dc6549095029ee578.pdf"

        gold_blocks_path = self.fixture_path / "4be952924cd565488b4a239dc6549095029ee578_lp_blocks.json"
        with open(gold_blocks_path, "r") as f:
            gold_blocks_json = json.load(f)
            gold_blocks_list = [Entity.from_json(entity_json=block) for block in gold_blocks_json]

        doc = self.parser.parse(input_pdf_path=input_pdf_path)
        images = self.rasterizer.rasterize(input_pdf_path=input_pdf_path, dpi=72)
        # PDF2ImageRasterizer.attach_images(images=images, doc=doc)
        doc.annotate_images(images=images)
        pred_blocks_list = self.layout_predictor.predict(doc=doc)

        for pred_block_ents, gold_block_ents in zip(pred_blocks_list, gold_blocks_list):
            assert pred_block_ents.metadata["type"] == gold_block_ents.metadata["type"]
            for pred_box, gold_box in zip(pred_block_ents.boxes, gold_block_ents.boxes):
                for field in ["page", "w", "h", "l", "t"]:
                    self.assertAlmostEqual(getattr(pred_box, field), getattr(gold_box, field), places=3)



import json
from argparse import ArgumentParser
from functools import partial
from typing import Any, Dict, List

import torch
import tqdm
from necessary import necessary
from sklearn.metrics import classification_report

from papermage.magelib import Box, Document, Entity, Image, Metadata, Span
from papermage.parsers.grobid_parser import GrobidClient, GrobidFullParser
from papermage.predictors import HFBIOTaggerPredictor, IVILATokenClassificationPredictor
from papermage.predictors.base_predictor import BasePredictor

with necessary("datasets"):
    import datasets


FULL_PATH = (
    "/net/nfs2.s2-research/shannons/projects/2108_S2VL_Sampling/s2-paper-sampling-and-annotation/"
    "combined_pipelines/combined-pdfs"
)


def patch_entity_from_json():
    def from_json(cls, entity_json: dict) -> "Entity":
        # the .get(..., None) or [] pattern is to handle the case where the key is present but the value is None
        return cls(
            spans=[Span.from_json(span_json=span_json) for span_json in entity_json.get("spans", None) or []],
            boxes=[Box.from_json(box_json=box_json) for box_json in entity_json.get("boxes", None) or []],
            metadata=Metadata.from_json(entity_json.get("metadata", None) or {}),
        )

    Entity.from_json = classmethod(from_json)


def patch_grobid_parser_xml_coords_to_boxes():
    old_xml_coords_to_boxes = GrobidFullParser._xml_coords_to_boxes

    def _xml_coords_to_boxes(self, coords_attribute: str, page_sizes: dict) -> List[Box]:
        current_page = getattr(self, "__current_page__", None)
        boxes = old_xml_coords_to_boxes(self=self, coords_attribute=coords_attribute, page_sizes=page_sizes)
        if current_page is not None:
            boxes = [box for box in boxes if box.page == current_page]
            for box in boxes:
                # when evaluating on a single page, the page number is always 0
                box.page = 0
        return boxes

    GrobidFullParser._xml_coords_to_boxes = _xml_coords_to_boxes


def patch_grobid_client_process_pdf():
    old_process_pdf = GrobidClient.process_pdf
    cache: Dict[str, Any] = {}

    def process_pdf(self, *args, pdf_file: str, **kwargs):
        if pdf_file not in cache:
            cache[pdf_file] = old_process_pdf(self=self, *args, pdf_file=pdf_file, **kwargs)
        return cache[pdf_file]

    GrobidClient.process_pdf = process_pdf


def run_vila(doc: Document, vila_predictor: BasePredictor) -> Document:
    entities = vila_predictor.predict(doc=doc)
    doc.annotate_entity(entities=entities, field_name="vila_entities")
    return doc


def run_grobid(doc: Document, grobid_parser: GrobidFullParser) -> Document:
    pdf_path = f"{FULL_PATH}/{doc.metadata.sha}.pdf"
    setattr(grobid_parser, "__current_page__", doc.metadata.page)
    grobid_doc = grobid_parser.parse(doc=doc, input_pdf_path=pdf_path)
    delattr(grobid_parser, "__current_page__")
    return grobid_doc


patch_entity_from_json()
patch_grobid_parser_xml_coords_to_boxes()
patch_grobid_client_process_pdf()

if __name__ == "__main__":

    ap = ArgumentParser()
    ap.add_argument("mode", choices=["new", "old", "grobid-fast", "grobid-full"])
    args = ap.parse_args()

    dt = datasets.load_dataset('allenai/s2-vl', split='test')
    device = "cuda" if torch.cuda.is_available() else "cpu"

    if args.mode == "new":
        vila_predictor = HFBIOTaggerPredictor.from_pretrained(
            model_name_or_path="allenai/vila-roberta-large-s2vl-internal",
            entity_name="tokens",
            context_name="pages",
            device=device,
        )
        run_fn = partial(run_vila, vila_predictor=vila_predictor)
    elif args.mode == "old":
        vila_predictor = IVILATokenClassificationPredictor.from_pretrained(
            "allenai/ivila-row-layoutlm-finetuned-s2vl-v2", device=device
        )
        run_fn = partial(run_vila, vila_predictor=vila_predictor)
    elif args.mode.startswith("grobid"):

        if args.mode == "grobid-full":
            grobid_parser = GrobidFullParser(grobid_server='http://s2-elanding-24.reviz.ai2.in:32771')
        else:
            grobid_parser = GrobidFullParser(grobid_server='http://s2-elanding-24.reviz.ai2.in:32772')
        run_fn = partial(run_grobid, grobid_parser=grobid_parser)
    else:
        raise ValueError(f"Invalid value for `mode`: {args.mode}")

    docs = []
    gold_tokens = []
    pred_tokens = []

    for row in tqdm.tqdm(dt, desc="Predicting", unit="doc"):
        doc = Document.from_json(row["doc"])
        images = [Image.from_base64(image) for image in row["images"]]
        doc.annotate_images(images=images)
        docs.append(doc)
        doc = run_fn(doc=doc)

        gold_tokens.extend(e[0].metadata.type if len(e := token._vila_entities)
                           else "null" for token in doc.tokens)
        pred_tokens.extend(e[0].metadata.label if len(e := token.vila_entities)
                           else "null" for token in doc.tokens)

    print(classification_report(y_true=gold_tokens, y_pred=pred_tokens, digits=4))



"""
@kylel
"""

from abc import abstractmethod

from papermage.magelib import Document


class Recipe:
    @abstractmethod
    def from_path(self, pdfpath: str) -> Document:
        raise NotImplementedError

    @abstractmethod
    def from_doc(self, doc: Document) -> Document:
        raise NotImplementedError



from papermage.predictors.api_predictors.span_qa_predictor import APISpanQAPredictor
from papermage.predictors.hf_predictors.bio_tagger_predictor import HFBIOTaggerPredictor
from papermage.predictors.hf_predictors.vila_predictor import (
    IVILATokenClassificationPredictor,
)
from papermage.predictors.hf_predictors.whitespace_predictor import WhitespacePredictor
from papermage.predictors.lp_predictors.block_predictor import LPBlockPredictor
from papermage.predictors.sklearn_predictors.word_predictor import SVMWordPredictor
from papermage.predictors.spacy_predictors.sentence_predictor import (
    PysbdSentencePredictor,
)

__all__ = [
    "HFBIOTaggerPredictor",
    "APISpanQAPredictor",
    "LPBlockPredictor",
    "IVILATokenClassificationPredictor",
    "WhitespacePredictor",
    "SVMWordPredictor",
    "PysbdSentencePredictor",
]



import json
import os
import pathlib
import unittest

from papermage.magelib import Document, Entity, Metadata, Span
from papermage.predictors.api_predictors.span_qa_predictor import APISpanQAPredictor


class TestSpanQAPredictor(unittest.TestCase):
    def setUp(self):
        self.fixture_path = pathlib.Path(__file__).parent.parent / "fixtures"
        with open(self.fixture_path / "2304.02623v1.json", "r") as f:
            test_doc_json = json.load(f)
        self.doc = Document.from_json(doc_json=test_doc_json)

        user_selected_span = Entity(
            spans=[Span(start=2784, end=2803)], metadata=Metadata(question="What does this mean?")
        )
        self.doc.annotate_entity(field_name="user_selected_span", entities=[user_selected_span])

        self.span_qa_predictor = APISpanQAPredictor(
            context_unit_name="rows",
        )

        # Should throw an error if the GPT4 prompt hasn't been cached
        self.span_qa_predictor.retrieval_qa_step.model.cache.enforce_cached = True

        self.using_github_actions = (
            "USING_GITHUB_ACTIONS" in os.environ and os.environ["USING_GITHUB_ACTIONS"] == "true"
        )

    def test_preprocess(self):
        paper_snippet = self.span_qa_predictor.preprocess(self.doc)
        assert paper_snippet.snippet == "AI-assisted writing"
        assert paper_snippet.qae[0].question == "What does this mean?"

    def test_retrieval(self):
        if self.using_github_actions:
            self.skipTest(
                "Skipping test_retrieval because it requires downloading and running a huggingface model."
            )
        paper_snippet = self.span_qa_predictor.preprocess(self.doc)

        self.span_qa_predictor.retrieval_qa_step.retrieve(paper_snippet)

        ev = [ev.paragraph for ev in paper_snippet.qae[0].evidence]
        assert ev == [
            "knowledge or information. For example, this could be researchers",
            "Position-Sensitive Definitions of Terms and Symbols. Proceedings of the 2021 CHI Conference on Human Factors in Computing Systems (2020).",
            "[1] Griffin Adams, Emily Alsentzer, Mert Ketenci, Jason Zucker, and Noémie El- hadad. 2021. What’s in a Summary? Laying the Groundwork for Advances in",
        ]

    def test_predict(self):
        if self.using_github_actions:
            self.skipTest("Skipping test_retrieval because it requires an openai key.")

        annotations = self.span_qa_predictor.predict(doc=self.doc)

        assert annotations[0].metadata["type"] == "answer"
        assert annotations[0].metadata["answer"] == 'The term "AI-assisted writing" refers to the utilization of Language Models (LLMs) to aid in writing tasks, as specified in the provided text.'
        assert annotations[0].metadata["question"] == "What does this mean?"

        self.doc.annotate_entity(field_name="context_with_span", entities=[annotations[1]])
        assert annotations[1].metadata["type"] == "context_with_span"
        assert annotations[1].text == "1 In the remainder of this work, unless otherwise specified, AI-assisted writing refers to the use of LLMs to support writing."

        self.doc.annotate_entity(field_name="evidence", entities=annotations[2:5])
        assert annotations[2].metadata["type"] == "evidence"
        assert annotations[2].text == "knowledge or information. For example, this could be researchers"

        assert annotations[3].metadata["type"] == "evidence"
        assert annotations[3].text == "Position-Sensitive Definitions of Terms and Symbols. Proceedings of the 2021 CHI Conference on Human Factors in Computing Systems (2020)."

        assert annotations[4].metadata["type"] == "evidence"
        assert annotations[4].text == "[1] Griffin Adams, Emily Alsentzer, Mert Ketenci, Jason Zucker, and Noémie El- hadad. 2021. What’s in a Summary? Laying the Groundwork for Advances in"



"""

Monkey patch the PIL.Image methods to add base64 conversion

@kylel, adapted some prior code from @shannons

"""

import base64
import os
from io import BytesIO

import numpy as np
from PIL import Image as pilimage_module
from PIL import ImageChops as pilimagechops_module
from PIL.Image import Image as PILImageClass


class Image:
    def __init__(self):
        self._pilimage = None

    @property
    def pilimage(self) -> PILImageClass:
        if not self._pilimage:
            raise AttributeError(f"This Image missing a PILImage. Try setting it first.")
        return self._pilimage

    @pilimage.setter
    def pilimage(self, pilimage: PILImageClass):
        if self._pilimage:
            raise AttributeError(f"This Image already has a PILImage. Make a new Image.")
        self._pilimage = pilimage

    @property
    def mode(self) -> str:
        return self.pilimage.mode

    def convert_to_greyscale(self) -> "Image":
        image = Image()
        image.pilimage = self.pilimage.convert("L").convert("RGB")
        return image

    def to_array(self) -> np.ndarray:
        return np.array(self.pilimage)

    @classmethod
    def from_array(cls, imarray: np.ndarray) -> "Image":
        if len(imarray.shape) != 3:
            raise ValueError(f"Input `imarray` should have 3 dimensions: (rows, cols, channels)")
        if imarray.shape[-1] != 3:
            raise ValueError(f"Input `imarray` final dimension should be length 3 for RGB.")
        my_image = Image()
        my_image._pilimage = pilimage_module.fromarray(imarray.astype("uint8"))
        return my_image

    def to_json(self):
        pass

    @classmethod
    def from_json(cls, image_json):
        pass

    def to_base64(self) -> str:
        # Ref: https://stackoverflow.com/a/31826470
        buffered = BytesIO()
        self.pilimage.save(buffered, format="PNG")
        img_str = base64.b64encode(buffered.getvalue())
        return img_str.decode("utf-8")

    @classmethod
    def from_base64(cls, img_str: str) -> "Image":
        buffered = BytesIO(base64.b64decode(img_str))
        im = pilimage_module.open(buffered)
        image = Image()
        image.pilimage = im
        return image

    @classmethod
    def create_rgb_all_white(cls, width: int = 600, height: int = 800) -> "Image":
        im = pilimage_module.new("RGB", (width, height), (255, 255, 255))
        image = Image()
        image.pilimage = im
        return image

    @classmethod
    def create_rgb_random(cls, width: int = 600, height: int = 800) -> "Image":
        imarray = np.random.rand(height, width, 3) * 255
        image = Image.from_array(imarray=imarray)
        return image

    def save(self, filepath: str, is_overwrite: bool = False, format: str = None, **params):
        if os.path.exists(filepath) and not is_overwrite:
            raise FileExistsError(f"Already image at {filepath}. Try setting `is_overwrite`")
        self.pilimage.save(filepath, format, **params)

    @classmethod
    def open(cls, filepath: str, mode: str = "r", formats: str = None) -> "Image":
        im = pilimage_module.open(filepath, mode, formats)
        image = Image()
        image.pilimage = im
        return image

    def show(self, title: str = None):
        self.pilimage.show(title)

    def __eq__(self, other: "Image") -> bool:
        diff = pilimagechops_module.difference(image1=self.pilimage, image2=other.pilimage)
        if diff.getbbox():
            return False
        else:
            return True

    def _repr_png_(self):
        """iPython display hook support for PNG format.

        :returns: PNG version of the image as bytes
        """
        return self.pilimage._repr_image("PNG", compress_level=1)

    def _repr_jpeg_(self):
        """iPython display hook support for JPEG format.

        :returns: JPEG version of the image as bytes
        """
        return self.pilimage._repr_image("JPEG")



"""

Sentence Splitter using PySBD

@shannons, @kylel

"""

import itertools
from typing import List, Tuple

import numpy as np
import pysbd

from papermage.magelib import (
    Document,
    Entity,
    PagesFieldName,
    Span,
    TokensFieldName,
    WordsFieldName,
    Annotation
)
from papermage.predictors.base_predictor import BasePredictor


def merge_neighbor_spans(spans: List[Span], distance=1) -> List[Span]:
    """Merge neighboring spans in a list of un-overlapped spans:
    when the gaps between neighboring spans is not larger than the
    specified distance, they are considered as the neighbors.

    Args:
        spans (List[Span]): The input list of spans.
        distance (int, optional):
            The upper bound of interval gaps between two neighboring spans.
            Defaults to 1.

    Returns:
        List[Span]: A list of merged spans
    """

    is_neighboring_spans = (
        lambda span1, span2: min(abs(span1.start - span2.end), abs(span1.end - span2.start)) <= distance
    )

    # It assumes non-overlapped intervals within the list
    def merge_neighboring_spans(span1, span2):
        return Span(min(span1.start, span2.start), max(span1.end, span2.end))

    spans = sorted(spans, key=lambda ele: ele.start)
    # When sorted, only one iteration round is needed.

    if len(spans) == 0:
        return []
    if len(spans) == 1:
        return spans

    cur_merged_spans = [spans[0]]

    for cur_span in spans[1:]:
        prev_span = cur_merged_spans.pop()
        if is_neighboring_spans(cur_span, prev_span):
            cur_merged_spans.append(merge_neighboring_spans(prev_span, cur_span))
        else:
            # In this case, the prev_span should be moved to the
            # bottom of the stack
            cur_merged_spans.extend([prev_span, cur_span])

    return cur_merged_spans


class PysbdSentencePredictor(BasePredictor):
    """Sentence Boundary based on Pysbd

    Examples:
        >>> doc: Document = parser.parse("path/to/pdf")
        >>> predictor = PysbdSentenceBoundaryPredictor()
        >>> sentence_spans = predictor.predict(doc)
        >>> doc.annotate(sentences=sentence_spans)
    """

    REQUIRED_BACKENDS = ["pysbd"]
    REQUIRED_DOCUMENT_FIELDS = [TokensFieldName]  # type: ignore

    def __init__(self) -> None:
        self._segmenter = pysbd.Segmenter(language="en", clean=False, char_span=True)

    def split_token_based_on_sentences_boundary(self, words: List[str]) -> List[Tuple[int, int]]:
        """
        Split a list of words into a list of (start, end) indices, indicating
        the start and end of each sentence.
        Duplicate of https://github.com/allenai/VILA/\blob/dd242d2fcbc5fdcf05013174acadb2dc896a28c3/src/vila/dataset/preprocessors/layout_indicator.py#L14      # noqa: E501

        Returns: List[Tuple(int, int)]
            a list of (start, end) for token indices within each sentence
        """

        if len(words) == 0:
            return [(0, 0)]
        combined_words = " ".join(words)

        char2token_mask = np.zeros(len(combined_words), dtype=np.int64)

        acc_word_len = 0
        for idx, word in enumerate(words):
            word_len = len(word) + 1
            char2token_mask[acc_word_len : acc_word_len + word_len] = idx
            acc_word_len += word_len

        segmented_sentences = self._segmenter.segment(combined_words)
        sent_boundary = [(ele.start, ele.end) for ele in segmented_sentences]

        split = []
        token_id_start = 0
        for start, end in sent_boundary:
            token_id_end = char2token_mask[start:end].max()
            if end + 1 >= len(char2token_mask) or char2token_mask[end + 1] != token_id_end:
                token_id_end += 1  # (Including the end)
            split.append((token_id_start, token_id_end))
            token_id_start = token_id_end
        return split

    def _predict(self, doc: Document) -> List[Annotation]:
        if hasattr(doc, WordsFieldName):
            words = [word.text for word in getattr(doc, WordsFieldName)]
            attr_name = WordsFieldName
            # `words` is preferred as it should has better reading
            # orders and text representation
        else:
            words = [token.text for token in doc.tokens]
            attr_name = TokensFieldName

        split = self.split_token_based_on_sentences_boundary(words)

        sentence_spans: List[Annotation] = []
        for start, end in split:
            if end - start == 0:
                continue
            if end - start < 0:
                raise ValueError

            cur_spans = getattr(doc, attr_name)[start:end]

            all_token_spans = list(itertools.chain.from_iterable([ele.spans for ele in cur_spans]))

            sentence_spans.append(Entity(spans=merge_neighbor_spans(all_token_spans)))

        return sentence_spans



"""

Base class for Predictors.

@shannons, @kylel

"""

from abc import abstractmethod
from typing import Any, Dict, List, Union

from papermage.magelib import Annotation, Document


class BasePredictor:
    @property
    @abstractmethod
    def REQUIRED_BACKENDS(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def REQUIRED_DOCUMENT_FIELDS(self):
        """Due to the dynamic nature of the document class as well the
        models, we require the model creator to provide a list of required
        fields in the document class. If not None, the predictor class
        will perform the check to ensure that the document contains all
        the specified fields.
        """
        raise NotImplementedError

    def _doc_field_checker(self, doc: Document) -> None:
        if self.REQUIRED_DOCUMENT_FIELDS is not None:
            for field in self.REQUIRED_DOCUMENT_FIELDS:
                assert (
                    field in doc.fields
                ), f"The input Document object {doc} doesn't contain the required field {field}"

    def predict(self, doc: Document) -> List[Annotation]:
        """For all the predictors, the input is a document object, and
        the output is a list of annotations.
        """
        self._doc_field_checker(doc)
        return self._predict(doc=doc)

    @abstractmethod
    def _predict(self, doc: Document) -> List[Annotation]:
        raise NotImplementedError



"""

SVM Word Predictor

Given a list of tokens, predict which tokens were originally part of the same word.
This does this in two phases: First, it uses a whitespace tokenizer to inform
whether tokens were originally part of the same word. Second, it uses a SVM
classifier to predict whether hyphenated segments should be considered a single word.

@kylel, @amanpreets

"""

import logging
import os
import re
import tarfile
import tempfile
from collections import defaultdict
from typing import Dict, List, Set, Tuple, Union
from urllib.parse import urlparse

import requests
from joblib import load
from scipy.sparse import hstack

from papermage.magelib import Document, Entity, Metadata, Span, WordsFieldName, Annotation
from papermage.parsers.pdfplumber_parser import PDFPlumberParser
from papermage.predictors.base_predictor import BasePredictor
from papermage.predictors.hf_predictors.whitespace_predictor import WhitespacePredictor


logger = logging.getLogger(__name__)


def make_text(entity: Union[Entity, Annotation], document: Document, field: str = WordsFieldName) -> str:
    candidate_words = document.find_by_span(entity, field)
    candidate_text: List[str] = []

    for i in range(len(candidate_words)):
        candidate_text.append(str(candidate_words[i].text))
        if i < len(candidate_words) - 1:
            next_word_start = candidate_words[i + 1].start
            curr_word_end = candidate_words[i].end
            assert isinstance(next_word_start, int), f"{candidate_words[i + 1]} has no span (non-int start)"
            assert isinstance(curr_word_end, int), f"{candidate_words[i]} has no span (non-int end)"
            if curr_word_end != next_word_start:
                candidate_text.append(document.symbols[curr_word_end : next_word_start])

    return "".join(candidate_text)


class IsWordResult:
    def __init__(self, original: str, new: str, is_edit: bool) -> None:
        self.original = original
        self.new = new
        self.is_edit = is_edit


class SVMClassifier:
    def __init__(self, ohe_encoder, scaler, estimator, unigram_probs):
        self.ohe_encoder = ohe_encoder
        self.scaler = scaler
        self.estimator = estimator
        self.unigram_probs = unigram_probs
        self.default_prob = unigram_probs["<unk>"]
        self.sparse_columns = [
            "shape",
            "s_bg1",
            "s_bg2",
            "s_bg3",
            "p_bg1",
            "p_bg2",
            "p_bg3",
            "p_lower",
            "s_lower",
        ]
        self.dense_columns = [
            "p_upper",
            "s_upper",
            "p_number",
            "s_number",
            "p_isalpha",
            "s_isalpha",
            "p_len",
            "s_len",
            "multi_hyphen",
            "uni_prob",
        ]

    @classmethod
    def from_path(cls, tar_path: str):
        with tempfile.TemporaryDirectory() as tmp_dir:
            if urlparse(url=tar_path).scheme != "":
                r = requests.get(tar_path)
                with open(os.path.join(tmp_dir, "svm_word_predictor.tar.gz"), "wb") as f:
                    f.write(r.content)
                tar_path = os.path.join(tmp_dir, "svm_word_predictor.tar.gz")
            with tarfile.open(tar_path, "r:gz") as tar:
                tar.extractall(path=tmp_dir)
                return cls.from_directory(tmp_dir)

    @classmethod
    def from_directory(cls, dir: str):
        classifier = SVMClassifier.from_paths(
            ohe_encoder_path=os.path.join(dir, "svm_word_predictor/ohencoder.joblib"),
            scaler_path=os.path.join(dir, "svm_word_predictor/scaler.joblib"),
            estimator_path=os.path.join(dir, "svm_word_predictor/hyphen_clf.joblib"),
            unigram_probs_path=os.path.join(dir, "svm_word_predictor/unigram_probs.pkl"),
        )
        return classifier

    @classmethod
    def from_paths(
        cls,
        ohe_encoder_path: str,
        scaler_path: str,
        estimator_path: str,
        unigram_probs_path: str,
    ):
        ohe_encoder = load(ohe_encoder_path)
        scaler = load(scaler_path)
        estimator = load(estimator_path)
        unigram_probs = load(unigram_probs_path)
        classifier = SVMClassifier(
            ohe_encoder=ohe_encoder,
            scaler=scaler,
            estimator=estimator,
            unigram_probs=unigram_probs,
        )
        return classifier

    def batch_predict(self, words: List[str], threshold: float) -> List[IsWordResult]:
        if any([word.startswith("-") or word.endswith("-") for word in words]):
            raise ValueError("Words should not start or end with hyphens.")

        all_features, word_id_to_feature_ids = self._get_features(words)
        all_scores = self.estimator.decision_function(all_features)
        results = []
        for word_id, feature_ids in word_id_to_feature_ids.items():
            word = words[word_id]
            word_segments = word.split("-")
            score_per_hyphen_in_word = all_scores[feature_ids]
            new_word = word_segments[0]
            for word_segment, hyphen_score in zip(word_segments[1:], score_per_hyphen_in_word):
                if hyphen_score > threshold:
                    new_word += "-"
                else:
                    new_word += ""
                new_word += word_segment
            results.append(IsWordResult(original=word, new=new_word, is_edit=word != new_word))
        return results

    def _get_dense_features(self, part: str, name_prefix: str):
        upper = int(part[0].isupper())
        number = int(part.isnumeric())
        alpha = int(part.isalpha())
        lower = part.lower()
        plen = len(part)
        return {
            f"{name_prefix}_upper": upper,
            f"{name_prefix}_number": number,
            f"{name_prefix}_isalpha": alpha,
            f"{name_prefix}_len": plen,
        }

    def _get_features(self, words: List[str]):
        sparse_all, dense_all = [], []
        idx, word_id_to_feature_ids = 0, dict()
        for widx, word in enumerate(words):
            split = word.split("-")
            for i, s in enumerate(split[:-1]):
                sparse_features, dense_features = dict(), dict()
                prefix = "-".join(split[: i + 1])
                suffix = "-".join(split[i + 1 :])
                if widx not in word_id_to_feature_ids:
                    word_id_to_feature_ids[widx] = []
                word_id_to_feature_ids[widx].append(idx)
                idx += 1
                dense_features.update(self._get_dense_features(prefix, "p"))
                dense_features.update(self._get_dense_features(suffix, "s"))
                orig_uni_prob = self.unigram_probs.get(word, self.default_prob)
                presuf_uni_prob = self.unigram_probs.get(f"{prefix}{suffix}", self.default_prob)
                dense_features["uni_prob"] = orig_uni_prob - presuf_uni_prob
                dense_features["multi_hyphen"] = int(word.count("-") > 1)
                sparse_features["shape"] = re.sub("\w", "x", word)
                sparse_features["s_lower"] = suffix.lower()
                sparse_features["s_bg1"] = suffix[:2]
                sparse_features["s_bg2"] = suffix[1:3] if len(suffix) > 2 else ""
                sparse_features["s_bg3"] = suffix[2:4] if len(suffix) > 3 else ""
                sparse_features["p_lower"] = prefix.lower()
                sparse_features["p_bg1"] = prefix[-2:][::-1] if len(prefix) > 1 else ""
                sparse_features["p_bg2"] = prefix[-3:-1][::-1] if len(prefix) > 2 else ""
                sparse_features["p_bg3"] = prefix[-4:-2][::-1] if len(prefix) > 3 else ""
                sparse_all.append([sparse_features[k] for k in self.sparse_columns])
                dense_all.append([dense_features[k] for k in self.dense_columns])
        dense_transformed = self.scaler.transform(dense_all)
        sparse_transformed = self.ohe_encoder.transform(sparse_all)

        return hstack([sparse_transformed, dense_transformed]), word_id_to_feature_ids


class SVMWordPredictor(BasePredictor):
    def __init__(
        self,
        classifier: SVMClassifier,
        threshold: float = -1.5,
        punct_as_words: str = PDFPlumberParser.DEFAULT_PUNCTUATION_CHARS.replace("-", ""),
    ):
        self.classifier = classifier
        self.whitespace_predictor = WhitespacePredictor()
        self.threshold = threshold
        self.punct_as_words = punct_as_words

    @classmethod
    def from_path(cls, tar_path: str):
        classifier = SVMClassifier.from_path(tar_path=tar_path)
        predictor = SVMWordPredictor(classifier=classifier)
        return predictor

    @classmethod
    def from_directory(cls, dir: str, threshold: float = -1.5):
        classifier = SVMClassifier.from_directory(dir=dir)
        predictor = SVMWordPredictor(classifier=classifier, threshold=threshold)
        return predictor

    def predict(self, doc: Document) -> List[Entity]:
        # clean input
        doc = self._make_clean_document(doc=doc)

        # validate input
        self._validate_tokenization(doc=doc)

        # initialize output data using whitespace tokenization
        # also avoid grouping specific punctuation. each instance should be their own word.
        (
            token_id_to_word_id,
            word_id_to_token_ids,
            word_id_to_text,
        ) = self._predict_with_whitespace(doc=doc)
        self._validate_token_word_assignments(word_id_to_token_ids=word_id_to_token_ids)

        # split up words back into tokens based on punctuation
        (
            token_id_to_word_id,
            word_id_to_token_ids,
            word_id_to_text,
        ) = self._keep_punct_as_words(
            doc=doc,
            word_id_to_token_ids=word_id_to_token_ids,
            punct_as_words=self.punct_as_words,
        )
        self._validate_token_word_assignments(word_id_to_token_ids=word_id_to_token_ids)

        # get hyphen word candidates
        hyphen_word_candidates = self._find_hyphen_word_candidates(
            tokens=doc.tokens,
            token_id_to_word_id=token_id_to_word_id,
            word_id_to_token_ids=word_id_to_token_ids,
            word_id_to_text=word_id_to_text,
        )
        candidate_texts = [
            word_id_to_text[prefix_word_id] + word_id_to_text[suffix_word_id]
            for prefix_word_id, suffix_word_id in hyphen_word_candidates
        ]

        # filter candidate texts
        hyphen_word_candidates, candidate_texts = self._filter_word_candidates(
            hyphen_word_candidates=hyphen_word_candidates,
            candidate_texts=candidate_texts,
        )

        # only triggers if there are hyphen word candidates
        if hyphen_word_candidates:
            # classify hyphen words
            results = self.classifier.batch_predict(words=candidate_texts, threshold=self.threshold)

            # update output data based on hyphen word candidates
            # first, we concatenate words based on prefix + suffix. this includes hyphen.
            # second, we modify the text value (e.g. remove hyphens) if classifier says.
            for (prefix_word_id, suffix_word_id), result in zip(hyphen_word_candidates, results):
                impacted_token_ids = word_id_to_token_ids[prefix_word_id] + word_id_to_token_ids[suffix_word_id]
                word_id_to_token_ids[prefix_word_id] = impacted_token_ids
                word_id_to_token_ids.pop(suffix_word_id)
                word_id_to_text[prefix_word_id] += word_id_to_text[suffix_word_id]
                word_id_to_text.pop(suffix_word_id)
                if result.is_edit is True:
                    word_id_to_text[prefix_word_id] = result.new
            token_id_to_word_id = {
                token_id: word_id for word_id, token_ids in word_id_to_token_ids.items() for token_id in token_ids
            }
            self._validate_token_word_assignments(word_id_to_token_ids=word_id_to_token_ids)

        # make into spangroups
        words = self._create_words(
            doc=doc,
            token_id_to_word_id=token_id_to_word_id,
            word_id_to_text=word_id_to_text,
        )
        return words

    def _make_clean_document(self, doc: Document) -> Document:
        """Word predictor doesnt work on documents with poor tokenizations,
        such as when there are empty tokens. This cleans up the document.
        We keep this pretty minimal, such as ignoring Span Boxes since dont need
        them for word prediction."""
        new_tokens = []
        for token in doc.tokens:
            if token.text.strip() != "":
                new_token = Entity(spans=[Span(start=span.start, end=span.end) for span in token.spans])
                new_tokens.append(new_token)
        new_doc = Document(symbols=doc.symbols)
        new_doc.annotate_entity(field_name="tokens", entities=new_tokens)
        return new_doc

    def _recursively_remove_trailing_hyphens(self, word: str) -> str:
        if word.endswith("-"):
            return self._recursively_remove_trailing_hyphens(word=word[:-1])
        else:
            return word

    def _validate_tokenization(self, doc: Document):
        """This Word Predictor relies on a specific type of Tokenization
        in which hyphens ('-') must be their own token. This verifies.

        Additionally, doesnt work unless there's an `.id` field on each token.
        See `_cluster_tokens_by_whitespace()` for more info.
        """
        for token in doc.tokens:
            if "-" in token.text and token.text != "-":
                raise ValueError(f"Document contains Token with hyphen, but not as its own token.")
            if token.id is None:
                raise ValueError(
                    f"Document contains Token without an `.id` field, which is necessary for this word Predictor's whitespace clustering operation."
                )

            if token.text.strip() == "":
                raise ValueError(f"Document contains Token with empty text, which is not allowed.")

    def _validate_token_word_assignments(self, word_id_to_token_ids, allow_missed_tokens: bool = True):
        for word_id, token_ids in word_id_to_token_ids.items():
            start = min(token_ids)
            end = max(token_ids)
            assert len(token_ids) == end - start + 1, f"word={word_id} comprised of disjoint token_ids={token_ids}"

        if not allow_missed_tokens:
            all_token_ids = {token_id for token_id in word_id_to_token_ids.values() for token_id in token_ids}
            if len(all_token_ids) < max(all_token_ids):
                raise ValueError(f"Not all tokens are assigned to a word.")

    def _cluster_tokens_by_whitespace(self, doc: Document) -> List[List[int]]:
        """
        `whitespace_tokenization` is necessary because lack of whitespace is an indicator that
        adjacent tokens belong in a word together.
        """
        _ws_tokens: List[Entity] = self.whitespace_predictor.predict(doc=doc)
        doc.annotate_entity(field_name="_ws_tokens", entities=_ws_tokens)

        # token -> ws_tokens
        token_id_to_ws_token_id = {}
        for token in doc.tokens:
            overlap_ws_tokens = token._ws_tokens
            if overlap_ws_tokens:
                token_id_to_ws_token_id[token.id] = overlap_ws_tokens[0].id

        # ws_token -> tokens
        ws_token_id_to_tokens = defaultdict(list)
        for token_id, ws_token_id in token_id_to_ws_token_id.items():
            ws_token_id_to_tokens[ws_token_id].append(token_id)

        # cluster tokens by whitespace
        clusters = []
        for ws_token_id, tokens in ws_token_id_to_tokens.items():
            clusters.append(sorted(tokens))

        # cleanup
        doc.remove_entity(field_name="_ws_tokens")
        return clusters

    def _predict_with_whitespace(self, doc: Document):
        """Predicts word boundaries using whitespace tokenization."""
        # precompute whitespace tokenization
        whitespace_clusters = self._cluster_tokens_by_whitespace(doc=doc)
        # assign word ids
        token_id_to_word_id = {}
        word_id_to_token_ids = defaultdict(list)
        for word_id, token_ids_in_cluster in enumerate(whitespace_clusters):
            for token_id in token_ids_in_cluster:
                token_id_to_word_id[token_id] = word_id
                word_id_to_token_ids[word_id].append(token_id)
        # get word strings
        word_id_to_text = {}
        for word_id, token_ids in word_id_to_token_ids.items():
            word_id_to_text[word_id] = "".join(doc.tokens[token_id].text for token_id in token_ids)
        return token_id_to_word_id, word_id_to_token_ids, word_id_to_text

    def _keep_punct_as_words(self, doc: Document, word_id_to_token_ids: Dict, punct_as_words: str):
        # keep track of which tokens are punctuation
        token_ids_are_punct = set()
        for token_id, token in enumerate(doc.tokens):
            if token.text in punct_as_words:
                token_ids_are_punct.add(token_id)
        # re-cluster punctuation tokens into their own words
        new_clusters = []
        for old_cluster in word_id_to_token_ids.values():
            for new_cluster in self._group_adjacent_with_exceptions(
                adjacent=old_cluster, exception_ids=token_ids_are_punct
            ):
                new_clusters.append(new_cluster)
        # reorder
        new_clusters = sorted(new_clusters, key=lambda x: min(x))
        # reassign word ids
        new_token_id_to_word_id = {}
        new_word_id_to_token_ids = defaultdict(list)
        for word_id, token_ids_in_cluster in enumerate(new_clusters):
            for token_id in token_ids_in_cluster:
                new_token_id_to_word_id[token_id] = word_id
                new_word_id_to_token_ids[word_id].append(token_id)
        # get word strings
        new_word_id_to_text = {}
        for word_id, token_ids in new_word_id_to_token_ids.items():
            new_word_id_to_text[word_id] = "".join(doc.tokens[token_id].text for token_id in token_ids)
        return new_token_id_to_word_id, new_word_id_to_token_ids, new_word_id_to_text

    def _group_adjacent_with_exceptions(self, adjacent: List[int], exception_ids: Set[int]) -> List[List[int]]:
        result = []
        group = []
        for e in adjacent:
            if e in exception_ids:
                if group:
                    result.append(group)
                result.append([e])
                group = []
            else:
                group.append(e)
        if group:
            result.append(group)
        return result

    def _find_hyphen_word_candidates(
        self,
        tokens,
        token_id_to_word_id,
        word_id_to_token_ids,
        word_id_to_text,
    ) -> Tuple[int, int]:
        """Finds the IDs of hyphenated words (in prefix + suffix format)."""
        # get all hyphen tokens
        # TODO: can refine this further by restricting to only tokens at end of `rows`
        hyphen_token_ids = []
        for token_id, token in enumerate(tokens):
            if token.text == "-":
                hyphen_token_ids.append(token_id)
        # get words that contain hyphen token, but only at the end (i.e. broken word)
        # these form the `prefix` of a potential hyphenated word
        #
        # edge case: sometimes the prefix word is *just* a hyphen. this means the word
        # itself is actually just a hyphen (e.g. like in tables of results)
        # dont consider these as candidates
        prefix_word_ids = set()
        for hyphen_token_id in hyphen_token_ids:
            prefix_word_id = token_id_to_word_id[hyphen_token_id]
            prefix_word_text = word_id_to_text[prefix_word_id]
            if prefix_word_text.endswith("-") and not prefix_word_text == "-":
                prefix_word_ids.add(prefix_word_id)
        # get words right after the prefix (assumed words in reading order)
        # these form the `suffix` of a potential hyphenated word
        # together, a `prefix` and `suffix` form a candidate pair
        #
        # edge case: sometimes the token stream ends with a hyphenated
        # word. this means suffix_word_id wont exist. dont consider these
        word_id_pairs = []
        for prefix_word_id in prefix_word_ids:
            suffix_word_id = prefix_word_id + 1
            suffix_word_text = word_id_to_text.get(suffix_word_id)
            if suffix_word_text is None:
                continue
            word_id_pairs.append((prefix_word_id, suffix_word_id))
        return sorted(word_id_pairs)

    def _filter_word_candidates(self, hyphen_word_candidates: list, candidate_texts: list) -> tuple:
        hyphen_word_candidates_filtered = []
        candidate_texts_filtered = []
        for hyphen_word_candidate, candidate_text in zip(hyphen_word_candidates, candidate_texts):
            if candidate_text.endswith("-") or candidate_text.startswith("-"):
                continue
            else:
                hyphen_word_candidates_filtered.append(hyphen_word_candidate)
                candidate_texts_filtered.append(candidate_text)
        return hyphen_word_candidates_filtered, candidate_texts_filtered

    def _create_words(self, doc: Document, token_id_to_word_id, word_id_to_text) -> List[Entity]:
        words = []
        tokens_in_word = [doc.tokens[0]]
        current_word_id = 0
        new_word_id = 0
        for token_id in range(1, len(doc.tokens)):
            token = doc.tokens[token_id]
            word_id = token_id_to_word_id.get(token_id)
            if word_id is None:
                logger.debug(f"Token {token_id} has no word ID. Likely PDF Parser produced empty tokens.")
                continue
            if word_id == current_word_id:
                tokens_in_word.append(token)
            else:
                spans = [
                    Span.create_enclosing_span(spans=[span for token in tokens_in_word for span in token.spans])
                ]
                metadata = Metadata(text=word_id_to_text[current_word_id]) if len(tokens_in_word) > 1 else None
                word = Entity(
                    spans=spans,
                    metadata=metadata,
                )
                words.append(word)
                tokens_in_word = [token]
                current_word_id = word_id
                new_word_id += 1
        # last bit
        spans = [Span.create_enclosing_span(spans=[span for token in tokens_in_word for span in token.spans])]
        metadata = Metadata(text=word_id_to_text[current_word_id]) if len(tokens_in_word) > 1 else None
        word = Entity(
            spans=spans,
            metadata=metadata,
        )
        words.append(word)
        new_word_id += 1
        return words



"""

@benjaminn

"""

import json
import unittest
from pathlib import Path

import springs
import torch
import transformers

from papermage.magelib import Document, Entity, Span
from papermage.predictors import HFBIOTaggerPredictor
from papermage.trainers.bio_tagger_predictor_trainer import (
    HFBIOTaggerPredictorTrainConfig,
    HFBIOTaggerPredictorTrainer,
)

TEST_SCIBERT_WEIGHTS = "allenai/scibert_scivocab_uncased"


class TestEntityClassificationPredictorTrainer(unittest.TestCase):
    def setUp(self):
        transformers.set_seed(407)
        self.fixture_path = Path("tests/fixtures")
        with open(self.fixture_path / "entity_classification_predictor_test_doc_papermage.json", "r") as f:
            test_doc_json = json.load(f)
        self.doc = Document.from_json(doc_json=test_doc_json)
        ent1 = Entity(spans=[Span(start=86, end=456)])
        ent2 = Entity(spans=[Span(start=457, end=641)])
        self.doc.annotate_entity(field_name="bibs", entities=[ent1, ent2])

        self.predictor = HFBIOTaggerPredictor.from_pretrained(
            model_name_or_path=TEST_SCIBERT_WEIGHTS,
            entity_name="tokens",
            context_name="pages",
            num_labels=3,
            id2label={0: "O", 1: "B_Title", 2: "I_Title"},
            label2id={"O": 0, "B_Title": 1, "I_Title": 2},
        )

        config = springs.from_dataclass(HFBIOTaggerPredictorTrainConfig)
        config = springs.merge(
            config,
            springs.from_dict(
                {
                    "data_path": self.fixture_path / "predictor_training_docs_tiny.jsonl",
                    "label_field": "words_starting_with_td",
                    "max_steps": 1,
                    "seed": 407,
                }
            ),
        )
        self.trainer = HFBIOTaggerPredictorTrainer(
            predictor=self.predictor,
            config=config,
        )

    def test_train(self) -> None:
        if (self.trainer.CACHE_PATH / self.trainer.data_id / "inputs.pt").exists():
            (self.trainer.CACHE_PATH / self.trainer.data_id / "inputs.pt").unlink()

        # self.trainer.train(docs_path="tests/fixtures/predictor_training_docs.jsonl", annotations_entity_name="bibs")
        id2label = {0: "O", 1: "B-words_starting_with_td", 2: "I-words_starting_with_td"}
        label2id = {"O": 0, "B-words_starting_with_td": 1, "I-words_starting_with_td": 2}

        self.trainer.predictor.predictor.config.id2label = id2label
        self.trainer.predictor.predictor.config.label2id = label2id
        self.trainer.train(
            docs_path=self.fixture_path / "predictor_training_docs_tiny.jsonl",
            annotations_entity_names=["words_starting_with_td"],
        )

        # check that the cache file exists
        assert (self.trainer.CACHE_PATH / self.trainer.data_id / "inputs.pt").exists()

        # check that we can load in the trained model and run it on the test doc (`self.doc`)
        new_predictor = HFBIOTaggerPredictor.from_pretrained(
            model_name_or_path=self.trainer.config.default_root_dir / "checkpoints",
            entity_name=self.predictor.entity_name,
            context_name=self.predictor.context_name,
            id2label=id2label,
            label2id=label2id,
        )

        token_tags = new_predictor.predict(doc=self.doc)
        assert len(token_tags) == 20

    def test_preprocess(self):
        self.trainer.predictor.predictor.config.id2label = {
            0: "O",
            1: "B-words_starting_with_td",
            2: "I-words_starting_with_td",
        }
        self.trainer.predictor.predictor.config.label2id = {
            "O": 0,
            "B-words_starting_with_td": 1,
            "I-words_starting_with_td": 2,
        }
        preprocessed_batches = self.trainer.preprocess(
            docs_path=self.fixture_path / "predictor_training_docs_tiny.jsonl",
            labels_fields=["words_starting_with_td"],
        )

        gold_preprocessed_batches = torch.load(self.fixture_path / "preprocessed_training_docs_tiny.pt")
        for gold_batch, test_batch in zip(gold_preprocessed_batches, preprocessed_batches):
            for key in set(gold_batch.keys()) | set(test_batch.keys()):
                assert torch.allclose(gold_batch[key], test_batch[key])

        # test multi-word entities
        self.trainer.predictor.predictor.config.id2label = {
            0: "O",
            1: "B-multi_word_entity",
            2: "I-multi_word_entity",
        }
        self.trainer.predictor.predictor.config.label2id = {
            "O": 0,
            "B-multi_word_entity": 1,
            "I-multi_word_entity": 2,
        }
        preprocessed_batches_mwe = self.trainer.preprocess(
            docs_path=self.fixture_path / "predictor_training_docs_tiny.jsonl",
            labels_fields=["multi_word_entity"],
        )
        gold_preprocessed_batches_mwe = torch.load(self.fixture_path / "preprocessed_training_docs_tiny_mwe.pt")
        for gold_batch, test_batch in zip(gold_preprocessed_batches_mwe, preprocessed_batches_mwe):
            for key in set(gold_batch.keys()) | set(test_batch.keys()):
                assert torch.allclose(gold_batch[key], test_batch[key])

        # test multiple fields
        id2label = {
            0: "O",
            1: "B-words_starting_with_td",
            2: "I-words_starting_with_td",
            3: "B-multi_word_entity",
            4: "I-multi_word_entity",
        }
        label2id = {lab: id for id, lab in id2label.items()}
        self.trainer.predictor.predictor.config.id2label = id2label
        self.trainer.predictor.predictor.config.label2id = label2id
        preprocessed_batches_multifields = self.trainer.preprocess(
            docs_path=self.fixture_path / "predictor_training_docs_tiny.jsonl",
            labels_fields=["words_starting_with_td", "multi_word_entity"],
        )
        gold_preprocessed_batches_multifields = torch.load(
            self.fixture_path / "preprocessed_training_docs_tiny_multifields.pt"
        )
        for gold_batch, test_batch in zip(gold_preprocessed_batches_multifields, preprocessed_batches_multifields):
            for key in set(gold_batch.keys()) | set(gold_batch.keys()):
                assert torch.allclose(gold_batch[key], test_batch[key])

    def test_eval(self):
        if (self.trainer.CACHE_PATH / self.trainer.data_id / "test_inputs.pt").exists():
            (self.trainer.CACHE_PATH / self.trainer.data_id / "test_inputs.pt").unlink()
        transformers.set_seed(407)
        self.trainer.predictor.predictor.config.id2label = {
            0: "O",
            1: "B-multi_word_entity",
            2: "I-multi_word_entity",
        }
        self.trainer.predictor.predictor.config.label2id = {
            "O": 0,
            "B-multi_word_entity": 1,
            "I-multi_word_entity": 2,
        }
        self.trainer.eval(
            docs_path=self.fixture_path / "predictor_training_docs_tiny.jsonl",
            annotations_entity_names=["multi_word_entity"],
        )

        # Check that inputs are cached
        assert (self.trainer.CACHE_PATH / self.trainer.data_id / "test_inputs.pt").exists()

        # Check that the outputs are correct
        with open(self.trainer.config.default_root_dir / "results" / "results.json") as f:
            pred_results = json.load(f)

        with open(self.fixture_path / "eval_results.json") as f:
            gold_results = json.load(f)

        assert gold_results["y_gold"] == pred_results["y_gold"]
        assert gold_results["y_hat"] == pred_results["y_hat"]

    def save_and_load_checkpoint(self):
        pass



"""

@geli-gel, @amanpreet692, @soldni

"""
import json
import os
import re
import warnings
import xml.etree.ElementTree as et
from collections import defaultdict
from tempfile import NamedTemporaryFile
from typing import Any, Dict, List, Optional, Tuple, cast

import numpy as np
from grobid_client.grobid_client import GrobidClient

from papermage.magelib import (
    Document,
    Entity,
    Metadata,
    PagesFieldName,
    RowsFieldName,
    Span,
    TokensFieldName,
)
from papermage.magelib.box import Box
from papermage.magelib.span import MergeClusterSpans
from papermage.parsers.parser import Parser

REQUIRED_DOCUMENT_FIELDS = [PagesFieldName, RowsFieldName, TokensFieldName]
NS = {"tei": "http://www.tei-c.org/ns/1.0"}


GROBID_VILA_MAP = {
    # title has no coordinates, not recoverable
    "title": ("Title",),
    # author has no coordinates, but they can be recovered from combining 'persName'
    "author": ("Author",),
    # abstract has no coordinates, but they can be recovered from combining 's'
    "abstract": ("Abstract",),
    # keywords has no coordinates, not recoverable
    "keywords": ("Keywords",),
    # sections are easily available. compared to VILA, section number is extracted and
    # reported as metadata.
    "head": ("Section",),
    # paragraph has no coordinates, but they can be recovered from combining 's'
    "p": ("Paragraph",),
    # list has no coordinates, but they can be recovered from combining 'item'
    "list": ("List",),
    "biblStruct": ("Bibliography",),
    "formula": ("Equation",),
    # tables are enclosed into figures, need to check
    "figure": ("Figure",),
    "table": ("Table",),
    # figDesc is caption for both tables and figures
    "figDesc": ("Caption",),
    # header and footers get merged into notes fields, but these are kinda tricky to
    # recover because things like author affiliations are also in the notes fields.
    # also, no coordinates.
    # for footnotes, need to check if this is of type "foot"
    "note": (
        "Header",
        "Footer",
        "Footnote",
    ),
}


def find_contiguous_ones(array):
    # Add a sentinel value at the beginning/end
    array = np.concatenate([[0], array, [0]])

    # Find the indexes where the array changes
    diff_indices = np.where(np.diff(array) != 0)[0] + 1
    # Find the start and end indexes of zero spans
    zero_start_indices = diff_indices[:-1:2]
    zero_end_indices = diff_indices[1::2] - 1

    spans = list(zip(zero_start_indices, zero_end_indices))

    # Exclude the spans with no element
    spans = [(start - 1, end) for start, end in spans if end - start >= 0]

    return spans


class GrobidFullParser(Parser):
    """Grobid parser that uses Grobid python client to hit a running
    Grobid server and convert resulting grobid XML TEI coordinates into
    PaperMage Annotations to annotate an existing Document.

    Run a Grobid server (from https://grobid.readthedocs.io/en/latest/Grobid-docker/):
    > docker pull lfoppiano/grobid:0.7.2
    > docker run -t --rm -p 8070:8070 lfoppiano/grobid:0.7.2
    """

    def __init__(self, check_server: bool = True, **grobid_config: Any):
        self.grobid_config = {
            "grobid_server": "http://localhost:8070",
            "batch_size": 1000,
            "sleep_time": 5,
            "timeout": 60,
            "coordinates": sorted(set((*GROBID_VILA_MAP.keys(), "s", "ref", "body", "item", "persName"))),
            **grobid_config,
        }
        assert "coordinates" in self.grobid_config, "Grobid config must contain 'coordinates' key"

        with NamedTemporaryFile(mode="w", delete=False) as f:
            json.dump(self.grobid_config, f)
            config_path = f.name

        self.client = GrobidClient(config_path=config_path, check_server=check_server)

        os.remove(config_path)

    def parse(  # type: ignore
        self,
        input_pdf_path: str,
        doc: Document,
        xml_out_dir: Optional[str] = None
    ) -> Document:
        assert doc.symbols != ""
        for field in REQUIRED_DOCUMENT_FIELDS:
            assert field in doc.fields

        (_, _, xml) = self.client.process_pdf(
            service="processFulltextDocument",
            pdf_file=input_pdf_path,
            generateIDs=False,
            consolidate_header=False,
            consolidate_citations=False,
            include_raw_citations=False,
            include_raw_affiliations=False,
            tei_coordinates=True,
            segment_sentences=True,
        )
        assert xml is not None, "Grobid returned no XML"

        if xml_out_dir:
            os.makedirs(xml_out_dir, exist_ok=True)
            xmlfile = os.path.join(xml_out_dir, os.path.basename(input_pdf_path).replace(".pdf", ".xml"))
            with open(xmlfile, "w") as f_out:
                f_out.write(xml)

        self._parse_xml_onto_doc(xml=xml, doc=doc)

        for p in getattr(doc, "p", []):
            grobid_text_elems = [s.metadata["grobid_text"] for s in p.s]
            grobid_text = " ".join(filter(lambda text: isinstance(text, str), grobid_text_elems))
            p.metadata["grobid_text"] = grobid_text

        # add vila-like entities
        vila_entities = self._make_vila_groups(doc)
        doc.annotate_entity(entities=vila_entities, field_name="vila_entities")
        return doc

    def _make_spans_from_boxes(self, doc: Document, entity: Entity) -> List[Span]:
        tokens = [cast(Entity, t) for match in doc.find_by_box(entity, "tokens") for t in match.tokens]
        spans = MergeClusterSpans(sorted(set(s for t in tokens for s in t.spans), key=lambda x: x.start)).merge()
        return spans

    def _make_spans_from_boxes_if_not_found(self, doc: Document, entity: Entity) -> List[Span]:
        spans = [Span(start=s.start, end=s.end) for s in entity.spans]
        if not spans:
            spans = self._make_spans_from_boxes(doc, entity)
        return spans

    def _make_entities_of_type(
        self, doc: Document, entities: List[Entity], entity_type: str, id_offset: int = 0
    ) -> List[Entity]:
        entities = [
            Entity(
                spans=self._make_spans_from_boxes_if_not_found(doc=doc, entity=ent),
                boxes=[Box(l=b.l, t=b.t, w=b.w, h=b.h, page=b.page) for b in ent.boxes],
                metadata=Metadata(**ent.metadata.to_json(), label=entity_type, id=id_offset + i),
            )
            for i, ent in enumerate(entities)
        ]
        return entities

    def _update_reserved_positions(
        self, reserved_positions: np.ndarray, entities: List[Entity]
    ) -> Tuple[List[Entity], np.ndarray]:
        new_entities: List[Entity] = []
        for ent in entities:
            new_spans = []
            for span in ent.spans:
                already_reserved = reserved_positions[span.start : span.end]
                for start, end in find_contiguous_ones(~already_reserved):
                    new_spans.append(Span(start=start + span.start, end=end + span.start))
                reserved_positions[span.start : span.end] = True
            if new_spans:
                new_entities.append(Entity(spans=new_spans, boxes=ent.boxes, metadata=ent.metadata))
        return new_entities, reserved_positions

    def _make_vila_groups(self, doc: Document) -> List[Entity]:
        ents: List[Entity] = []
        reserved_positions = np.zeros(len(doc.symbols), dtype=bool)

        if _ := getattr(doc, "title", []):
            # title has no coordinates, so we can't recover its position!
            pass

        if h := getattr(doc, "author", []):
            h_ = self._make_entities_of_type(doc=doc, entities=h, entity_type="Author", id_offset=len(ents))
            h__, reserved_positions = self._update_reserved_positions(reserved_positions, h_)
            ents.extend(h__)

        if a := getattr(doc, "abstract", []):
            a_ = self._make_entities_of_type(doc=doc, entities=a, entity_type="Abstract", id_offset=len(ents))
            a__, reserved_positions = self._update_reserved_positions(reserved_positions, a_)
            ents.extend(a__)

        if _ := getattr(doc, "keywords", []):
            # keywords has no coordinates, so we can't recover their positions!
            pass

        if s := getattr(doc, "head", []):
            s_ = self._make_entities_of_type(doc=doc, entities=s, entity_type="Section", id_offset=len(ents))
            s__, reserved_positions = self._update_reserved_positions(reserved_positions, s_)
            ents.extend(s__)

        if t := getattr(doc, "list", []):
            t_ = self._make_entities_of_type(doc=doc, entities=t, entity_type="List", id_offset=len(ents))
            t__, reserved_positions = self._update_reserved_positions(reserved_positions, t_)
            ents.extend(t__)

        if b := getattr(doc, "biblStruct", []):
            b_ = self._make_entities_of_type(doc=doc, entities=b, entity_type="Bibliography", id_offset=len(ents))
            b__, reserved_positions = self._update_reserved_positions(reserved_positions, b_)
            ents.extend(b__)

        if e := getattr(doc, "formula", []):
            e_ = self._make_entities_of_type(doc=doc, entities=e, entity_type="Equation", id_offset=len(ents))
            e__, reserved_positions = self._update_reserved_positions(reserved_positions, e_)
            ents.extend(e__)

        if figs := getattr(doc, "figure", []):
            for fig in figs:
                current_boxes = [Box(l=b.l, t=b.t, w=b.w, h=b.h, page=b.page) for b in fig.boxes]

                if "figDesc" in doc.fields:
                    caption_boxes = [b for d in doc.find_by_box(fig, "figDesc") for b in d.boxes]
                    current_boxes = [b for b in current_boxes if b not in caption_boxes]

                if "table" in doc.fields:
                    table_boxes = [b for d in doc.find_by_box(fig, "table") for b in d.boxes]
                    current_boxes = [b for b in current_boxes if b not in table_boxes]

                if not current_boxes:
                    continue

                new_fig = Entity(
                    spans=self._make_spans_from_boxes(doc, Entity(boxes=current_boxes)),
                    boxes=current_boxes,
                    metadata=Metadata(**fig.metadata.to_json(), label="Figure", id=len(ents)),
                )
                new_figs, reserved_positions = self._update_reserved_positions(reserved_positions, [new_fig])
                ents.extend(new_figs)

        if t := getattr(doc, "table", []):
            t_ = self._make_entities_of_type(doc=doc, entities=t, entity_type="Table", id_offset=len(ents))
            t__, reserved_positions = self._update_reserved_positions(reserved_positions, t_)
            ents.extend(t__)

        if c := getattr(doc, "figDesc", []):
            c_ = self._make_entities_of_type(doc=doc, entities=c, entity_type="Caption", id_offset=len(ents))
            c__, reserved_positions = self._update_reserved_positions(reserved_positions, c_)
            ents.extend(c__)

        if _ := getattr(doc, "note", []):
            # notes have no coordinates, so we can't recover their positions!
            pass

        if p := getattr(doc, "p", []):
            p_ = self._make_entities_of_type(doc=doc, entities=p, entity_type="Paragraph", id_offset=len(ents))
            p__, reserved_positions = self._update_reserved_positions(reserved_positions, p_)
            ents.extend(p__)

        return ents

    def _parse_xml_onto_doc(self, xml: str, doc: Document) -> Document:
        try:
            xml_root = et.fromstring(xml)
        except Exception as e:
            if xml == "[GENERAL] An exception occurred while running Grobid.":
                warnings.warn("Grobid returned an error; check server logs")
                return doc
            raise e

        all_box_groups = self._get_box_groups(xml_root)
        for field, box_groups in all_box_groups.items():
            doc.annotate_entity(field_name=field, entities=box_groups)

        return doc

    def _xml_coords_to_boxes(self, coords_attribute: str, page_sizes: dict) -> List[Box]:
        coords_list = coords_attribute.split(";")
        boxes = []
        for coords in coords_list:
            if coords == "":
                # this page has no coordinates
                continue
            pg, x, y, w, h = coords.split(",")
            proper_page = int(pg) - 1
            boxes.append(
                Box(l=float(x), t=float(y), w=float(w), h=float(h), page=proper_page).to_relative(
                    *page_sizes[proper_page]
                )
            )
        return boxes

    def _get_box_groups(self, root: et.Element) -> Dict[str, List[Entity]]:
        page_size_root = root.find(".//tei:facsimile", NS)
        assert page_size_root is not None, "No facsimile found in Grobid XML"

        page_size_data = page_size_root.findall(".//tei:surface", NS)
        page_sizes = dict()
        for data in page_size_data:
            page_sizes[int(data.attrib["n"]) - 1] = [float(data.attrib["lrx"]), float(data.attrib["lry"])]

        all_boxes: Dict[str, List[Entity]] = defaultdict(list)

        for field in self.grobid_config["coordinates"]:
            structs = root.findall(f".//tei:{field}", NS)
            for i, struct in enumerate(structs):
                if (coords_str := struct.attrib.get("coords", None)) is None:
                    all_coords = struct.findall(".//*[@coords]")
                    coords_str = ";".join([c.attrib["coords"] for c in all_coords if "coords" in c.attrib])

                if coords_str == "":
                    continue

                if not (boxes := self._xml_coords_to_boxes(coords_str, page_sizes)):
                    # we check if the boxes are empty because sometimes users monkey-patch
                    # _xml_coords_to_boxes to filter by page
                    continue

                metadata_dict: Dict[str, Any] = {
                    f"grobid_{re.sub(r'[^a-zA-Z0-9_]+', '_', k)}": v
                    for k, v in struct.attrib.items()
                    if k != "coords"
                }
                metadata_dict["grobid_order"] = i
                metadata_dict["grobid_text"] = struct.text
                metadata = Metadata.from_json(metadata_dict)
                box_group = Entity(boxes=boxes, metadata=metadata)
                all_boxes[field].append(box_group)

        return all_boxes


if __name__ == "__main__":
    from argparse import ArgumentParser

    from papermage.parsers import PDFPlumberParser

    ap = ArgumentParser()
    ap.add_argument("pdf_path", type=str)
    opts = ap.parse_args()

    doc = PDFPlumberParser().parse(opts.pdf_path)
    doc = GrobidFullParser().parse(opts.pdf_path, doc)

    for p in getattr(doc, "p", []):
        for s in p.s:
            print(s.metadata.grobid_text)



"""

Tests for Metadata

@lucas

"""

import unittest
from copy import deepcopy

from papermage.magelib import Metadata


class TestMetadata(unittest.TestCase):
    def test_add_keys(self):
        metadata = Metadata()

        metadata["foo"] = 1
        self.assertEqual(metadata.foo, 1)

        metadata.bar = 2
        self.assertEqual(metadata.bar, 2)

        metadata.set("baz", 3)
        self.assertEqual(metadata.baz, 3)

    def test_access_keys(self):
        metadata = Metadata()
        metadata.foo = "bar"

        self.assertEqual(metadata.foo, "bar")
        self.assertEqual(metadata.get("foo"), "bar")
        self.assertTrue(metadata["foo"])
        self.assertIsNone(metadata.get("bar"))

    def test_json_transform(self):
        metadata = Metadata.from_json({"foo": "bar"})

        self.assertEqual(metadata.to_json(), {"foo": "bar"})
        self.assertEqual(Metadata.from_json(metadata.to_json()), metadata)

    def test_len(self):
        metadata = Metadata.from_json({f"k{i}": i for i in range(10)})
        self.assertEqual(len(metadata), 10)

        metadata.pop("k0")
        self.assertEqual(len(metadata), 9)

        del metadata.k1
        self.assertEqual(len(metadata), 8)

    def test_valid_names(self):
        metadata = Metadata()

        # this should work fine
        metadata.set("foo", "bar")
        self.assertEqual(metadata.foo, "bar")

        # this should fail because `1foo` is not a valid python variable name
        with self.assertRaises(ValueError):
            metadata.set("1foo", "bar")

    def test_deep_copy(self):
        metadata = Metadata.from_json({"foo": 1, "bar": 2, "baz": 3})
        metadata2 = deepcopy(metadata)
        self.assertEqual(metadata, metadata2)

    def test_get_unknown_key(self):
        metadata = Metadata()
        self.assertIsNone(metadata.text)



"""

A (start, end) interval that references some string. For example, span (0, 4)
represents the word 'This' in the string 'This is a document.'

@kylel, @egork

"""

from typing import List, Dict, List

from collections import defaultdict


class Span:
    __slots__ = ['start', 'end']

    def __init__(self, start: int, end: int):
        self.start = start
        self.end = end

    def to_json(self) -> List[int]:
        """Returns whatever representation is JSON compatible"""
        return [self.start, self.end]

    @classmethod
    def from_json(cls, span_json: List) -> "Span":
        """Recreates the object from the JSON serialization"""
        return Span(start=span_json[0], end=span_json[-1])

    def __repr__(self):
        return f'Span{self.to_json()}'

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Span):
            return False
        return self.start == other.start and self.end == other.end

    def __lt__(self, other: 'Span'):
        """Useful for sort(). Orders according to the start index.
        If ties, then order according to the end index."""
        if self.start == other.start:
            return self.end < other.end
        return self.start < other.start

    def __hash__(self) -> int:
        return hash((self.start, self.end))

    def is_overlap(self, other: 'Span') -> bool:
        """Whether self overlaps with the other Span object."""
        return (
            self.start <= other.start < self.end
            or other.start <= self.start < other.end
            or self == other
        )

    @classmethod
    def create_enclosing_span(cls, spans: List['Span']) -> 'Span':
        """Create the narrowest Span that completely encloses all the input Spans."""
        if not spans:
            raise ValueError(f'`spans` should be non-empty.')
        start = spans[0].start
        end = spans[0].end
        for span in spans[1:]:
            if span.start < start:
                start = span.start
            if span.end > end:
                end = span.end
        return Span(start=start, end=end)


class MergeClusterSpans:
    """
    Merge neighboring spans which are index distance apart
    Inspired by https://leetcode.com/problems/merge-intervals/

    Originally @egork, Revised @kylel
    """

    def __init__(
            self,
            spans: List[Span],
            index_distance: int = 1
    ) -> None:
        """
        Args
            index_distance (int): Distance between the spans
        """
        self._spans = spans
        self._index_distance = index_distance
        self._graph = self._build_graph(spans=spans, index_distance=index_distance)
        self._clusters = None

    @property
    def spans(self) -> List[Span]:
        return self._spans

    @property
    def index_distance(self) -> int:
        return self._index_distance

    @index_distance.setter
    def index_distance(self, d: int):
        """If modify this distance, everything that's been computed before
        should be recomputed."""
        self._index_distance = d
        self._graph = self._build_graph(spans=self.spans, index_distance=d)
        if self._clusters:
            self._clusters = self._cluster(spans=self.spans)

    @property
    def clusters(self) -> List[List[Span]]:
        if not self._clusters:
            self._clusters = self._cluster(spans=self.spans)
        return self._clusters

    @staticmethod
    def _is_neighboring_spans(span1: Span, span2: Span, index_distance: int) -> bool:
        """Whether two spans are considered neighboring"""
        return min(
            abs(span1.start - span2.end), abs(span1.end - span2.start)
        ) <= index_distance

    def _build_graph(self, spans: List[Span], index_distance: int) -> Dict[int, List[int]]:
        """
        Build graph, each node is the position within the input list of spans.
        Spans are considered overlapping if they are index_distance apart
        """
        graph = defaultdict(list)
        for i, span_i in enumerate(spans):
            for j in range(i + 1, len(spans)):
                if self._is_neighboring_spans(span1=span_i, span2=spans[j], index_distance=index_distance):
                    graph[i].append(j)
                    graph[j].append(i)
        return graph

    def _cluster(self, spans: List[Span]) -> List[List[Span]]:
        """Cluster nodes (i.e. spans) by finding connected components"""
        if len(spans) == 0:
            return [[]]

        visited = set()
        num_components = 0
        component_id_to_members = defaultdict(list)

        def _dfs(start: int):
            stack = [start]
            while stack:
                pos = stack.pop()
                if pos not in visited:
                    visited.add(pos)
                    component_id_to_members[num_components].append(pos)
                    stack.extend(self._graph[pos])

        # mark all nodes in the same connected component with the same integer.
        for i, span in enumerate(spans):
            if i not in visited:
                _dfs(start=i)
                num_components += 1

        return [
            [spans[member_id] for member_id in sorted(component_id_to_members[n])]
            for n in range(num_components)
        ]

    def merge(self) -> List[Span]:
        """
        For each of the lists of the connected nodes, merge into bigger Spans
        """
        merged_spans = []
        for cluster in self.clusters:
            if cluster:
                merged_span = Span.create_enclosing_span(spans=cluster)
                merged_spans.append(merged_span)
        return merged_spans
